#!/usr/bin/env python

"""
Create release notes for a new release of this GitHub repository.
"""

# Requires:
#
# * assumes current directory is within a repository clone
# * pyGithub (conda or pip install) - https://pygithub.readthedocs.io/
# * Github personal access token (https://github.com/settings/tokens)
#
# Github token access is needed or the GitHub API limit
# will likely interfere with making a complete report
# of the release.

import argparse
import datetime
import github
import logging
import os


logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("create_release_notes")


def findGitConfigFile():
    """
    return full path to .git/config file
    
    must be in current working directory or some parent directory
    
    This is a simplistic search that could be improved by using 
    an open source package.
    
    Needs testing for when things are wrong.
    """
    path = os.getcwd()
    for i in range(99):
        config_file = os.path.join(path, ".git", "config")
        if os.path.exists(config_file):
            return config_file  # found it!

        # next, look in the parent directory
        path = os.path.abspath(os.path.join(path, ".."))

    msg = "Could not find .git/config file in any parent directory."
    logger.error(msg)
    raise ValueError(msg)


def parse_git_url(url):
    """
    return (organization, repository) tuple from url line of .git/config file
    """
    if url.startswith("git@"):  # deal with git@github.com:org/repo.git
        url = url.split(":")[1]
    org, repo = url.rstrip(".git").split("/")[-2:]
    return org, repo


def getRepositoryInfo():
    """
    return (organization, repository) tuple from .git/config file
    
    This is a simplistic search that could be improved by using 
    an open source package.
    
    Needs testing for when things are wrong.
    """
    config_file = findGitConfigFile()

    with open(config_file, "r") as f:
        for line in f.readlines():
            line = line.strip()
            if line.startswith("url"):
                url = line.split("=")[-1].strip()
                if url.find("github.com") < 0:
                    msg = "Not a GitHub repo: " + url
                    logger.error(msg)
                    raise ValueError(msg)
                return parse_git_url(url)


def get_release_info(token, base_tag_name, head_branch_name, milestone_name):
    """mine the Github API for information about this release"""
    organization_name, repository_name = getRepositoryInfo()
    gh = github.Github(token)  # GitHub Personal Access Token

    user = gh.get_user(organization_name)
    logger.debug(f"user: {user}")

    repo = user.get_repo(repository_name)
    logger.debug(f"repo: {repo}")

    # fmt: off
    milestones = [
        m
        for m in repo.get_milestones(state="all")
        if m.title == milestone_name
    ]
    # fmt: on
    if len(milestones) == 0:
        msg = f"Could not find milestone: {milestone_name}"
        logger.error(msg)
        raise ValueError(msg)
    milestone = milestones[0]
    logger.debug(f"milestone: {milestone}")

    compare = repo.compare(base_tag_name, head_branch_name)
    logger.debug(f"compare: {compare}")

    commits = {c.sha: c for c in compare.commits}
    logger.debug(f"# commits: {len(commits)}")

    tags = {}
    earliest = None
    for t in repo.get_tags():
        if t.commit.sha in commits:
            tags[t.name] = t
        elif t.name == base_tag_name:
            # PyGitHub oddity:
            #   t.commit == commit
            #   t.commit.last_modified != commit.last_modified
            commit = repo.get_commit(t.commit.sha)
            dt = str2time(commit.last_modified)
            earliest = min(dt, earliest or dt)
    logger.debug(f"# tags: {len(tags)}")

    # fmt: off
    pulls = {
        p.number: p
        for p in repo.get_pulls(state="closed")
        if p.closed_at > earliest
    }
    # fmt: on
    logger.debug(f"# pulls: {len(pulls)}")

    # fmt: off
    issues = {
        i.number: i
        for i in repo.get_issues(milestone=milestone, state="closed")
        if (
            (milestone is not None or i.closed_at > earliest)
            and
            i.number not in pulls
        )
    }
    # fmt: on
    logger.debug(f"# issues: {len(issues)}")

    return repo, milestone, tags, pulls, issues, commits


def parse_command_line():
    """command line argument parser"""
    doc = __doc__.strip()
    parser = argparse.ArgumentParser(description=doc)

    help_text = "name of tag to start the range"
    parser.add_argument("base", action="store", help=help_text)

    help_text = "name of milestone"
    parser.add_argument("milestone", action="store", help=help_text)

    parser.add_argument(
        "token",
        action="store",
        # fmt: off
        help=(
            "personal access token "
            "(see: https://github.com/settings/tokens)"
        )
        # fmt: on
    )

    help_text = "name of tag, branch, SHA to end the range"
    help_text += ' (default="main")'
    # fmt: off
    parser.add_argument(
        "--head", 
        action='store', 
        dest='head',
        nargs='?', 
        help = help_text, 
        default="main"
    )
    # fmt: on

    return parser.parse_args()


def str2time(time_string):
    """convert date/time string to datetime object
    
    input string example: ``Tue, 20 Dec 2016 17:35:40 GMT``
    """
    if time_string is None:
        msg = f"need valid date/time string, not: {time_string}"
        logger.error(msg)
        raise ValueError(msg)
    return datetime.datetime.strptime(time_string, "%a, %d %b %Y %H:%M:%S %Z")


def report(title, repo, milestone, tags, pulls, issues, commits):
    print(f"## {title}")
    print("")
    print(f"* **date/time**: {datetime.datetime.now()}")
    print("* **release**: ")
    print("* **documentation**: [PDF]()")
    if milestone is not None:
        print(f"* **milestone**: [{milestone.title}]({milestone.url})")
        print("")
    print("section | quantity")
    print("-" * 5, " | ", "-" * 5)
    print(f"[New Tags](#tags) | {len(tags)}")
    print(f"[Pull Requests](#pull-requests) | {len(pulls)}")
    print(f"[Issues](#issues) | {len(issues)}")
    print(f"[Commits](#commits) | {len(commits)}")
    print("")
    print("### Tags")
    print("")
    if len(tags) == 0:
        print("-- none --")
    else:
        print("tag | date | name")
        print("-" * 5, " | ", "-" * 5, " | ", "-" * 5)
        for k, tag in sorted(tags.items()):
            commit = repo.get_commit(tag.commit.sha)
            when = str2time(commit.last_modified).strftime("%Y-%m-%d")
            print(f"[{tag.commit.sha[:7]}]({tag.commit.html_url}) | {when} | {k}")
    print("")
    print("### Pull Requests")
    print("")
    if len(pulls) == 0:
        print("-- none --")
    else:
        print("pull request | date | state | title")
        print("-" * 5, " | ", "-" * 5, " | ", "-" * 5, " | ", "-" * 5)
        for k, pull in sorted(pulls.items()):
            state = {True: "merged", False: "closed"}[pull.merged]
            when = str2time(pull.last_modified).strftime("%Y-%m-%d")
            # fmt: off
            print(
                f"[#{pull.number}]({pull.html_url})"
                f" | {when} | {state} | {pull.title}"
            )
            # fmt: on
    print("")
    print("### Issues")
    print("")
    if len(issues) == 0:
        print("-- none --")
    else:
        print("issue | date | title")
        print("-" * 5, " | ", "-" * 5, " | ", "-" * 5)
        for k, issue in sorted(issues.items()):
            if k not in pulls:
                when = issue.closed_at.strftime("%Y-%m-%d")
                # fmt: off
                print(
                    f"[#{issue.number}]({issue.html_url})"
                    f" | {when} | {issue.title}"
                )
                # fmt: on
    print("")
    print("### Commits")
    print("")
    if len(commits) == 0:
        print("-- none --")
    else:
        print("commit | date | message")
        print("-" * 5, " | ", "-" * 5, " | ", "-" * 5)
        for k, commit in commits.items():
            message = commit.commit.message.splitlines()[0]
            when = commit.raw_data["commit"]["committer"]["date"].split("T")[0]
            print(f"[{k[:7]}]({commit.html_url}) | {when} | {message}")


def main(base=None, head=None, milestone=None, token=None, debug=False):
    if debug:
        base_tag_name = base
        head_branch_name = head
        milestone_name = milestone
        logger.setLevel(logging.DEBUG)
    else:
        cmd = parse_command_line()
        base_tag_name = cmd.base
        head_branch_name = cmd.head
        milestone_name = cmd.milestone
        token = cmd.token
        logger.setLevel(logging.WARNING)

    # fmt: off
    info = get_release_info(
        token, base_tag_name, head_branch_name, milestone_name
    )
    # fmt: on
    # milestone, repo, tags, pulls, issues, commits = info
    report(milestone_name, *info)


if __name__ == "__main__":
    main()


# NeXus - Neutron and X-ray Common Data Format
#
# Copyright (C) 2008-2022 NeXus International Advisory Committee (NIAC)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For further information, see http://www.nexusformat.org



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Dict mapping custom schema instances from eln_data.yaml file on concepts in NXapm."""

NxApmElnInput = {"IGNORE": {"fun": "load_from_dict_list", "terms": "em_lab/detector"},
                 "IGNORE": {"fun": "load_from", "terms": "em_lab/ebeam_column/aberration_correction/applied"},
                 "IGNORE": {"fun": "load_from_dict_list", "terms": "em_lab/ebeam_column/aperture_em"},
                 "/ENTRY[entry*]/PROGRAM[program2]/program": {"fun": "load_from", "terms": "atom_probe/control_software_program"},
                 "/ENTRY[entry*]/PROGRAM[program2]/program/@version": {"fun": "load_from", "terms": "atom_probe/control_software_program__attr_version"},
                 "/ENTRY[entry*]/experiment_identifier": {"fun": "load_from", "terms": "entry/experiment_identifier"},
                 "/ENTRY[entry*]/start_time": {"fun": "load_from", "terms": "entry/start_time"},
                 "/ENTRY[entry*]/end_time": {"fun": "load_from", "terms": "entry/end_time"},
                 "/ENTRY[entry*]/run_number": {"fun": "load_from", "terms": "entry/run_number"},
                 "/ENTRY[entry*]/operation_mode": {"fun": "load_from", "terms": "entry/operation_mode"},
                 "/ENTRY[entry*]/experiment_description": {"fun": "load_from", "terms": "entry/experiment_description"},
                 "IGNORE": {"fun": "load_from", "terms": "sample/alias"},
                 "/ENTRY[entry*]/sample/grain_diameter": {"fun": "load_from", "terms": "sample/grain_diameter/value"},
                 "/ENTRY[entry*]/sample/grain_diameter/@units": {"fun": "load_from", "terms": "sample/grain_diameter/unit"},
                 "/ENTRY[entry*]/sample/grain_diameter_error": {"fun": "load_from", "terms": "sample/grain_diameter_error/value"},
                 "/ENTRY[entry*]/sample/grain_diameter_error/@units": {"fun": "load_from", "terms": "sample/grain_diameter_error/unit"},
                 "/ENTRY[entry*]/sample/heat_treatment_quenching_rate": {"fun": "load_from", "terms": "sample/heat_treatment_quenching_rate/value"},
                 "/ENTRY[entry*]/sample/heat_treatment_quenching_rate/@units": {"fun": "load_from", "terms": "sample/heat_treatment_quenching_rate/unit"},
                 "/ENTRY[entry*]/sample/heat_treatment_quenching_rate_error": {"fun": "load_from", "terms": "sample/heat_treatment_quenching_rate_error/value"},
                 "/ENTRY[entry*]/sample/heat_treatment_quenching_rate_error/@units": {"fun": "load_from", "terms": "sample/heat_treatment_quenching_rate_error/unit"},
                 "/ENTRY[entry*]/sample/heat_treatment_temperature": {"fun": "load_from", "terms": "sample/heat_treatment_temperature/value"},
                 "/ENTRY[entry*]/sample/heat_treatment_temperature/@units": {"fun": "load_from", "terms": "sample/heat_treatment_temperature/unit"},
                 "/ENTRY[entry*]/sample/heat_treatment_temperature_error": {"fun": "load_from", "terms": "sample/heat_treatment_temperature_error/value"},
                 "/ENTRY[entry*]/sample/heat_treatment_temperature_error/@units": {"fun": "load_from", "terms": "sample/heat_treatment_temperature_error/unit"},
                 "/ENTRY[entry*]/specimen/name": {"fun": "load_from", "terms": "specimen/name"},
                 "/ENTRY[entry*]/specimen/preparation_date": {"fun": "load_from", "terms": "specimen/preparation_date"},
                 "IGNORE": {"fun": "load_from", "terms": "specimen/sample_history"},
                 "/ENTRY[entry*]/specimen/alias": {"fun": "load_from", "terms": "specimen/alias"},
                 "/ENTRY[entry*]/specimen/is_polycrystalline": {"fun": "load_from", "terms": "specimen/is_polycrystalline"},
                 "/ENTRY[entry*]/specimen/description": {"fun": "load_from", "terms": "specimen/description"},
                 "/ENTRY[entry*]/atom_probe/FABRICATION[fabrication]/identifier": {"fun": "load_from", "terms": "atom_probe/fabrication_identifier"},
                 "/ENTRY[entry*]/atom_probe/FABRICATION[fabrication]/model": {"fun": "load_from", "terms": "atom_probe/fabrication_model"},
                 "/ENTRY[entry*]/atom_probe/FABRICATION[fabrication]/vendor": {"fun": "load_from", "terms": "atom_probe/fabrication_vendor"},
                 "/ENTRY[entry*]/atom_probe/analysis_chamber/pressure": {"fun": "load_from", "terms": "atom_probe/analysis_chamber_pressure/value"},
                 "/ENTRY[entry*]/atom_probe/analysis_chamber/pressure/@units": {"fun": "load_from", "terms": "atom_probe/analysis_chamber_pressure/unit"},
                 "/ENTRY[entry*]/atom_probe/control_software/PROGRAM[program1]/program": {"fun": "load_from", "terms": "atom_probe/control_software_program"},
                 "/ENTRY[entry*]/atom_probe/control_software/PROGRAM[program1]/program/@version": {"fun": "load_from", "terms": "atom_probe/control_software_program__attr_version"},
                 "/ENTRY[entry*]/atom_probe/field_of_view": {"fun": "load_from", "terms": "atom_probe/field_of_view/value"},
                 "/ENTRY[entry*]/atom_probe/field_of_view/@units": {"fun": "load_from", "terms": "atom_probe/field_of_view/unit"},
                 "/ENTRY[entry*]/atom_probe/flight_path_length": {"fun": "load_from", "terms": "atom_probe/flight_path_length/value"},
                 "/ENTRY[entry*]/atom_probe/flight_path_length/@units": {"fun": "load_from", "terms": "atom_probe/flight_path_length/unit"},
                 "/ENTRY[entry*]/atom_probe/instrument_name": {"fun": "load_from", "terms": "atom_probe/instrument_name"},
                 "/ENTRY[entry*]/atom_probe/ion_detector/model": {"fun": "load_from", "terms": "atom_probe/ion_detector_model"},
                 "/ENTRY[entry*]/atom_probe/ion_detector/name": {"fun": "load_from", "terms": "atom_probe/ion_detector_name"},
                 "/ENTRY[entry*]/atom_probe/ion_detector/serial_number": {"fun": "load_from", "terms": "atom_probe/ion_detector_serial_number"},
                 "/ENTRY[entry*]/atom_probe/ion_detector/type": {"fun": "load_from", "terms": "atom_probe/ion_detector_type"},
                 "/ENTRY[entry*]/atom_probe/local_electrode/name": {"fun": "load_from", "terms": "atom_probe/local_electrode_name"},
                 "/ENTRY[entry*]/atom_probe/location": {"fun": "load_from", "terms": "atom_probe/location"},
                 "/ENTRY[entry*]/atom_probe/REFLECTRON[reflectron]/applied": {"fun": "load_from", "terms": "atom_probe/reflectron_applied"},
                 "/ENTRY[entry*]/atom_probe/stage_lab/base_temperature": {"fun": "load_from", "terms": "atom_probe/stage_lab_base_temperature/value"},
                 "/ENTRY[entry*]/atom_probe/stage_lab/base_temperature/@units": {"fun": "load_from", "terms": "atom_probe/stage_lab_base_temperature/unit"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/detection_rate": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_detection_rate/value"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/detection_rate/@units": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_detection_rate/unit"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/initial_radius": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_initial_radius/value"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/initial_radius/@units": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_initial_radius/unit"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/shank_angle": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_shank_angle/value"},
                 "/ENTRY[entry*]/atom_probe/specimen_monitoring/shank_angle/@units": {"fun": "load_from", "terms": "atom_probe/specimen_monitoring_shank_angle/unit"},
                 "/ENTRY[entry*]/atom_probe/status": {"fun": "load_from", "terms": "atom_probe/status"},
                 "/ENTRY[entry*]/atom_probe/pulser/pulse_fraction": {"fun": "load_from", "terms": "atom_probe/pulser/pulse_fraction"},
                 "/ENTRY[entry*]/atom_probe/pulser/pulse_frequency": {"fun": "load_from", "terms": "atom_probe/pulser/pulse_frequency/value"},
                 "/ENTRY[entry*]/atom_probe/pulser/pulse_frequency/@units": {"fun": "load_from", "terms": "atom_probe/pulser/pulse_frequency/unit"},
                 "/ENTRY[entry*]/atom_probe/pulser/pulse_mode": {"fun": "load_from", "terms": "atom_probe/pulser/pulse_mode"},
                 "/ENTRY[entry*]/atom_probe/ranging/PROGRAM[program1]/program": {"fun": "load_from", "terms": "atom_probe/ranging/program"},
                 "/ENTRY[entry*]/atom_probe/ranging/PROGRAM[program1]/program/@version": {"fun": "load_from", "terms": "atom_probe/ranging/program__attr_version"},
                 "/ENTRY[entry*]/atom_probe/reconstruction/PROGRAM[program1]/program": {"fun": "load_from", "terms": "atom_probe/reconstruction/program"},
                 "/ENTRY[entry*]/atom_probe/reconstruction/PROGRAM[program1]/program/@version": {"fun": "load_from", "terms": "atom_probe/reconstruction/program__attr_version"},
                 "/ENTRY[entry*]/atom_probe/reconstruction/crystallographic_calibration": {"fun": "load_from", "terms": "atom_probe/reconstruction/crystallographic_calibration"},
                 "/ENTRY[entry*]/atom_probe/reconstruction/parameter": {"fun": "load_from", "terms": "atom_probe/reconstruction/parameter"},
                 "/ENTRY[entry*]/atom_probe/reconstruction/protocol_name": {"fun": "load_from", "terms": "atom_probe/reconstruction/protocol_name"}}

# NeXus concept specific mapping tables which require special treatment as the current
# NOMAD OASIS custom schema implementation delivers them as a list of dictionaries instead
# of a directly flattenable list of keyword, value pairs

NxUserFromListOfDict = {"/ENTRY[entry*]/USER[user*]/name": {"fun": "load_from", "terms": "name"},
                        "/ENTRY[entry*]/USER[user*]/affiliation": {"fun": "load_from", "terms": "affiliation"},
                        "/ENTRY[entry*]/USER[user*]/address": {"fun": "load_from", "terms": "address"},
                        "/ENTRY[entry*]/USER[user*]/email": {"fun": "load_from", "terms": "email"},
                        "/ENTRY[entry*]/USER[user*]/orcid": {"fun": "load_from", "terms": "orcid"},
                        "/ENTRY[entry*]/USER[user*]/orcid_platform": {"fun": "load_from", "terms": "orcid_platform"},
                        "/ENTRY[entry*]/USER[user*]/telephone_number": {"fun": "load_from", "terms": "telephone_number"},
                        "/ENTRY[entry*]/USER[user*]/role": {"fun": "load_from", "terms": "role"},
                        "/ENTRY[entry*]/USER[user*]/social_media_name": {"fun": "load_from", "terms": "social_media_name"},
                        "/ENTRY[entry*]/USER[user*]/social_media_platform": {"fun": "load_from", "terms": "social_media_platform"}}

# LEAP6000 can use up to two lasers and voltage pulsing (both at the same time?)
NxPulserFromListOfDict = {"/ENTRY[entry*]/atom_probe/pulser/SOURCE[source*]/name": {"fun": "load_from", "terms": "name"},
                          "/ENTRY[entry*]/atom_probe/pulser/SOURCE[source*]/power": {"fun": "load_from", "terms": "power"},
                          "/ENTRY[entry*]/atom_probe/pulser/SOURCE[source*]/pulse_energy": {"fun": "load_from", "terms": "pulse_energy"},
                          "/ENTRY[entry*]/atom_probe/pulser/SOURCE[source*]/wavelength": {"fun": "load_from", "terms": "wavelength"}}



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The abstract class off of which to implement readers."""
from abc import ABC, abstractmethod
from typing import Tuple, Any


class BaseReader(ABC):
    """
    The abstract class off of which to implement readers.

    The filename's prefix is the identifier. The '_reader.py' is snipped out.
    For this BaseReader with filename base_reader.py the ID  becomes 'base'

    For future reference:
    - Support links by setting the path in the template with the following object
       object = {"link": "/path/to/source/data"}
    """

    # pylint: disable=too-few-public-methods

    __name__ = "BaseReader"

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = [""]

    @abstractmethod
    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Reads data from given file and returns a filled template dictionary"""
        return template


READER = BaseReader



"""For functions that directly or indirectly help to for rendering ELN.
Note that this not schema eln that is rendered to Nomad rather the eln that
is generated by schema eln."""

# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re
from typing import Any, Dict
import xml.etree.ElementTree as ET
import yaml

from pynxtools.dataconverter.helpers import generate_template_from_nxdl
from pynxtools.dataconverter.template import Template
from pynxtools.nexus.nexus import get_nexus_definitions_path


def retrieve_nxdl_file(nexus_def: str) -> str:
    """Retrive full path of nexus file.

    Parameters
    ----------
    nexus_def : str
        Name of nexus definition e.g. NXmpes

    Returns
    -------
    str
        Returns full path of file e.g. <full_path>/NXmpes.nxdl.xml

    Raises
    ------
    ValueError
        Need correct definition name, e.g. NXmpes not NXmpes.nxdl.xml
    """
    definition_path = get_nexus_definitions_path()

    def_path = os.path.join(definition_path,
                            'contributed_definitions',
                            f"{nexus_def}.nxdl.xml")
    if os.path.exists(def_path):
        return def_path

    def_path = os.path.join(definition_path,
                            'base_definitions',
                            f"{nexus_def}.nxdl.xml")

    if os.path.exists(def_path):
        return def_path

    def_path = os.path.join(definition_path,
                            'applications',
                            f"{nexus_def}.nxdl.xml")
    if os.path.exists(def_path):
        return def_path

    raise ValueError("Incorrect definition is rendered, try with correct definition name.")


def get_empty_template(nexus_def: str) -> Template:
    """Generate eln in yaml file.

    Parameters
    ----------
    nexus_def : str
        Name of NeXus definition e.g. NXmpes

    Return
    ------
        Template
    """

    nxdl_file = retrieve_nxdl_file(nexus_def)
    nxdl_root = ET.parse(nxdl_file).getroot()
    template = Template()
    generate_template_from_nxdl(nxdl_root, template)

    return template


def take_care_of_special_concepts(key: str):
    """For some special concepts such as @units."""
    def unit_concept():
        return {'value': None,
                'unit': None}

    if key == '@units':
        return unit_concept()


def get_recursive_dict(concatenated_key: str,
                       recursive_dict: Dict[str, Any],
                       level_to_skip: int) -> None:
    """Get recursive dict for concatenated string of keys.

    Parameters
    ----------
    concatenated_key : str
        String of keys separated by slash
    recursive_dict : dict
        Dict to recursively stroring data.
    level_to_skip : int
        Integer to skip the level of hierarchical level
    """
    # splitig keys like: '/entry[ENTRY]/position[POSITION]/xx'.
    # skiping the first empty '' and top parts as directed by users.
    key_li = concatenated_key.split('/')[level_to_skip + 1:]
    # list of key for special consideration
    sp_key_li = ['@units']
    last_key = ""
    last_dict = {}
    for key in key_li:
        if '[' in key and '/' not in key:
            key = re.findall(r'\[(.*?)\]', key,)[0].capitalize()
        if not key:
            continue
        last_key = key
        last_dict = recursive_dict
        if key in recursive_dict:
            if recursive_dict[key] is None:
                recursive_dict[key] = {}
                recursive_dict = recursive_dict[key]

            else:
                if key in sp_key_li:
                    recursive_dict.update(take_care_of_special_concepts(key))
                else:
                    recursive_dict = recursive_dict[key]
        else:
            if key in sp_key_li:
                recursive_dict.update(take_care_of_special_concepts(key))
            else:
                recursive_dict[key] = {}
                recursive_dict = recursive_dict[key]
    # For special key cleaning parts occurs inside take_care_of_special_concepts func.
    if last_key not in sp_key_li:
        last_dict[last_key] = None


def generate_eln(nexus_def: str, eln_file: str = '', level_to_skip: int = 1) -> None:
    """Genrate eln from application definition.

    Parameters
    ----------
    nexus_def : str
        _description_
    eln_file : str
        _description_

    Returns:
        None
    """

    template = get_empty_template(nexus_def)
    recursive_dict: Dict[str, Any] = {}
    for key, _ in template.items():
        get_recursive_dict(key, recursive_dict, level_to_skip)

    name_split = eln_file.rsplit('.')
    if not eln_file:
        if nexus_def[0:2] == 'NX':
            raw_name = nexus_def[2:]
            eln_file = raw_name + '.yaml'

    elif len(name_split) == 1:
        eln_file = eln_file + '.yaml'

    elif len(name_split) == 2 and name_split[1] == 'yaml':
        pass
    else:
        raise ValueError("Eln file should come with 'yaml' extension or without extension.")

    with open(eln_file, encoding='utf-8', mode='w') as eln_f:
        yaml.dump(recursive_dict, sort_keys=False, stream=eln_f)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility class to analyze which vendor/community files are passed to em reader."""

# pylint: disable=no-member,duplicate-code

from typing import Tuple, Dict, List


class EmUseCaseSelector:  # pylint: disable=too-few-public-methods
    """Decision maker about what needs to be parsed given arbitrary input.

    Users might invoke this dataconverter with arbitrary input, no input, or
    too much input. The UseCaseSelector decide what to do in each case.
    """

    def __init__(self, file_paths: Tuple[str] = None):
        """Initialize the class.

        dataset injects numerical data and metadata from an analysis.
        eln injects additional metadata and eventually numerical data.
        """
        self.case: Dict[str, list] = {}
        self.vendor_parser = "none"
        self.vendor: List[str] = []
        self.eln: List[str] = []
        self.eln_parser = "none"
        self.is_valid = False
        self.supported_mime_types = ["bcf", "emd", "dm3", "h5oina", "yaml", "yml"]
        for mime_type in self.supported_mime_types:
            self.case[mime_type] = []
        for file_name in file_paths:
            index = file_name.lower().rfind(".")
            if index >= 0:
                suffix = file_name.lower()[index + 1::]
                add = (suffix in self.supported_mime_types) \
                    and (file_name not in self.case[suffix])
                if add is True:
                    self.case[suffix].append(file_name)
        # the em reader currently supports a combination of one vendor file and
        # one ELN/YAML file, vendor files can come from different microscopes
        # which requires to distinguish between which reader has to be used
        # the OxfordInstrument reader for H5OINA
        # or the HyperSpy reader for Bruker BCF, Velox EMD, or Digital Micrograph DM3
        oina_input = 0
        for mime_type, value in self.case.items():
            if mime_type in ["h5oina"]:
                oina_input += len(value)
        hspy_input = 0
        for mime_type, value in self.case.items():
            if mime_type in ["bcf", "dm3", "emd"]:
                hspy_input += len(value)

        assert (oina_input == 1) or (hspy_input == 1), \
            "Currently the reader supports to have only one vendor input file!"
        if oina_input == 1:
            self.vendor += self.case["h5oina"]
            self.vendor_parser = "oina"
        if hspy_input == 1:
            for mime_type in ["bcf", "emd", "dm3"]:
                self.vendor += self.case[mime_type]
                self.vendor_parser = "hspy"

        eln_input = len(self.case["yaml"]) + len(self.case["yml"])
        assert eln_input == 1, \
            "Currently the reader supports to have only one YAML input-file!"

        for mime_type in ["yaml", "yml"]:
            self.eln += self.case[mime_type]
            self.eln_parser = "nomad-oasis"

        self.is_valid = True

# test = EmUseCaseSelector(("a.bcf", "b.yaml", "c.apt", "d.h5oina"))



"""NeXus classes are defined in the NeXus Definition Language (NXDL).

The NeXus Definition Language is an XML Schema Definition (XSD).
Each NeXus class is defined in an XML file using the NXDL schema.
"""

from .discover import find_definition  # noqa F401
from .discover import iter_definitions  # noqa F401
from .syntax import nxdl_schema  # noqa F401
from .syntax import validate_definition  # noqa F401



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""A dispersion reader for reading dispersion data from a rii yaml file"""
from typing import List, Any, Dict
import re
from datetime import datetime
import numpy as np
import pandas as pd
import yaml

from pynxtools.dataconverter.readers.rii_database.nx_data_writer import NXdataWriter
from pynxtools.dataconverter.readers.rii_database.citation_parser import Citation
import pynxtools.dataconverter.readers.rii_database.dispersion_functions as dispersion


class DispersionReader:
    """Reads a rii dispersion from a yml file"""

    def __init__(self, download_bibtex: bool = False):
        self.entries: Dict[str, Any] = {}
        self.single_params: Dict[str, Dict[str, float]] = {}
        self.repeated_params: Dict[str, Dict[str, np.ndarray]] = {}
        self.download_bibtex = download_bibtex

    def add_table(self, path: str, nx_name: str, tabular_data: pd.DataFrame):
        """Adds a tabular dataframe to the nexus entries dict"""
        disp_path = f"{path}/DISPERSION_TABLE[{nx_name}]"

        if f"{disp_path}/refractive_index" in self.entries:
            if not np.array_equal(
                self.entries[f"{disp_path}/wavelength"], tabular_data.index.values
            ):
                raise ValueError(
                    "Trying to add tabular dispersions with different wavelength ranges"
                )
            self.entries[f"{disp_path}/refractive_index"] += tabular_data[
                "refractive_index"
            ].values
            return

        self.entries[f"{path}/model_name"] = "Table"
        self.entries[f"{disp_path}/model_name"] = "Table"
        self.entries[f"{disp_path}/convention"] = "n + ik"
        self.entries[f"{disp_path}/wavelength"] = tabular_data.index.values
        self.entries[f"{disp_path}/wavelength/@units"] = "micrometer"
        self.entries[f"{disp_path}/refractive_index"] = tabular_data[
            "refractive_index"
        ].values

    def add_model_name(self, path: str, name: str):
        """Adds a model name to the nexus entries dict"""
        self.entries[f"{path}/model_name"] = name

    def add_model_info(
        self,
        dispersion_info: dispersion.DispersionInfo,
        model_input: dispersion.ModelInput,
    ):
        """Adds a model info the the nexus entries dict"""
        disp_path = (
            f"{model_input.dispersion_path}/"
            f"DISPERSION_FUNCTION[{dispersion_info.name.lower()}]"
        )

        if model_input.wavelength_range:
            self.add_wavelength_bounds(disp_path, model_input.wavelength_range)

        self.entries[f"{disp_path}/model_name"] = dispersion_info.name
        representation = dispersion_info.representation
        formula = dispersion_info.formula
        if dispersion_info.representation == "eps" and model_input.convert_to_n:
            representation = "n"
            formula = f"sqrt({dispersion_info.formula})"
        self.entries[f"{disp_path}/formula"] = f"{representation} = {formula}"
        self.entries[f"{disp_path}/representation"] = representation
        self.entries[f"{disp_path}/convention"] = "n + ik"
        self.entries[f"{disp_path}/wavelength_identifier"] = "lambda"
        self.entries[f"{disp_path}/wavelength_unit"] = 1
        self.entries[f"{disp_path}/wavelength_unit/@units"] = "micrometer"

        return disp_path

    def add_single_param(self, path: str, name: str, value: float, unit: str):
        """
        Adds a single parameter to the nexus entries dict and the internal
        single_params dict.
        """
        self.entries[f"{path}/DISPERSION_SINGLE_PARAMETER[{name}]/name"] = name
        self.entries[f"{path}/DISPERSION_SINGLE_PARAMETER[{name}]/value"] = value
        self.entries[f"{path}/DISPERSION_SINGLE_PARAMETER[{name}]/value/@units"] = unit

        single_param_path = self.single_params.setdefault(path, {})
        single_param_path[name] = value

    def add_rep_param(
        self, path: str, name: str, values: List[float], units: List[str]
    ):
        """
        Adds a repeated parameter to the nexus entries dict and the internal
        repeated_params dict
        """
        if not units:
            raise ValueError("Units must be specified")

        self.entries[f"{path}/DISPERSION_REPEATED_PARAMETER[{name}]/name"] = name
        self.entries[f"{path}/DISPERSION_REPEATED_PARAMETER[{name}]/values"] = values
        if len(units) > 1:
            self.entries[
                f"{path}/DISPERSION_REPEATED_PARAMETER[{name}]/parameter_units"
            ] = units
        self.entries[
            f"{path}/DISPERSION_REPEATED_PARAMETER[{name}]/values/@units"
        ] = units[0]

        repeated_param_path = self.repeated_params.setdefault(path, {})
        repeated_param_path[name] = np.array(values)

    def add_wavelength_bounds(self, path: str, bound_str: str):
        """Adds upper and lower wavelength bounds to the nexus entries dict"""
        lower, upper = list(map(float, bound_str.split()))
        self.entries[f"{path}/wavelength_min"] = lower
        self.entries[f"{path}/wavelength_min/@units"] = "micrometer"
        self.entries[f"{path}/wavelength_max"] = upper
        self.entries[f"{path}/wavelength_max/@units"] = "micrometer"

    def read_metadata(self, yml_file: dict, path: str):
        """Reads metadata from a dispersion yaml file"""
        refs = [
            Citation(
                'M. N. Polyanskiy, "Refractive index database," '
                '<a href="https://refractiveindex.info">https://refractiveindex.info<\\a>. '
                f"Accessed on {datetime.today().strftime('%Y-%m-%d')}.",
                desc="This entry is part of the refractiveindex.info database.",
            )
        ]
        refs[0].bibtex = (
            "@misc{rii,\n"
            "    author = {Mikhail N. Polyanskiy},\n"
            "    title = {Refractive index database},\n"
            "    howpublished = {\\url{https://refractiveindex.info}},\n"
            f"    note = {{Accessed on {datetime.today().strftime('%Y-%m-%d')}}}\n"
            "}"
        )
        if "REFERENCES" in yml_file:
            refs += Citation.parse_citations(
                yml_file["REFERENCES"], download_bibtex=self.download_bibtex
            )

        for ref in refs:
            ref.write_entries(self.entries, path)

        if "COMMENTS" in yml_file:
            self.entries[f"{path}/refractiveindex_info_comment"] = yml_file["COMMENTS"]

            temp = re.search(r"(\d+(?:\.\d*)?)\s+(℃|°C|℉|°F|K)", yml_file["COMMENTS"])
            if temp:
                clean_unit = temp.group(2).replace("℃", "°C").replace("℉", "°F")
                self.entries["/ENTRY[entry]/sample/temperature"] = temp.group(1)
                self.entries["/ENTRY[entry]/sample/temperature/@units"] = clean_unit

    def write_dispersions(
        self,
        yml_file: dict,
        table_input: dispersion.TableInput,
        model_input: dispersion.ModelInput,
    ):
        """Writes all dispersions from the dict into the entries dict"""

        def write_dispersion(
            dispersion_relation: Dict[str, Any],
            table_input: dispersion.TableInput,
            model_input: dispersion.ModelInput,
        ):
            """Writes a given dispersion relation into the entries dict"""

            type_to_nx_name = {
                "tabulated k": "dispersion_table_k",
                "tabulated n": "dispersion_table_n",
                "tabulated nk": "dispersion_table_nk",
            }
            if dispersion_relation["type"] in dispersion.SUPPORTED_TABULAR_PARSERS:
                table_input.table_type = dispersion_relation["type"]
                table_input.data = dispersion_relation["data"]
                table_input.nx_name = type_to_nx_name[dispersion_relation["type"]]
                dispersion.tabulated(table_input)
                return

            if dispersion_relation["type"] in dispersion.FORMULA_PARSERS:
                coeffs = list(map(float, dispersion_relation["coefficients"].split()))
                model_input.coeffs = coeffs
                model_input.wavelength_range = dispersion_relation.get(
                    "wavelength_range", None
                )
                dispersion.FORMULA_PARSERS[dispersion_relation["type"]](model_input)
                return

            raise NotImplementedError(
                f'No parser for type {dispersion_relation["type"]}'
            )

        def is_table_k_formula_n_disp(dispersion1: dict, dispersion2: dict):
            """Checks if the two dispersions are one tabular k formula and one formula n"""
            if "tabulated k" == dispersion1["type"] and dispersion2["type"].startswith(
                "formula"
            ):
                return True

            if "tabulated k" == dispersion2["type"] and dispersion1["type"].startswith(
                "formula"
            ):
                return True

            return False

        if len(yml_file["DATA"]) == 2 and is_table_k_formula_n_disp(*yml_file["DATA"]):
            model_input.convert_to_n = True

        for dispersion_relation in yml_file["DATA"]:
            write_dispersion(dispersion_relation, table_input, model_input)

    def read_dispersion(
        self,
        filename: str,
        identifier: str = "dispersion_x",
    ):
        """Reads a dispersion from a yml input file"""
        with open(filename, "r", encoding="utf-8") as yml:
            yml_file = yaml.load(
                yml,
                yaml.SafeLoader,
            )
        dispersion_path = f"/ENTRY[entry]/{identifier}"

        # Only read metadata for the ordinary axis
        # as this should be the same for all axes
        if identifier == "dispersion_x":
            self.read_metadata(yml_file, "/ENTRY[entry]")

        self.write_dispersions(
            yml_file,
            dispersion.TableInput(dispersion_path, "", "", "", self.add_table),
            dispersion.ModelInput(
                dispersion_path,
                [],
                False,
                dispersion.TemplateCallbacks(
                    self.add_model_name,
                    self.add_model_info,
                    self.add_single_param,
                    self.add_rep_param,
                ),
            ),
        )

        self.entries = NXdataWriter(
            self.entries, dispersion_path, self.single_params, self.repeated_params
        ).write_entries()

        return self.entries



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse H5OINA and create IPF color map default plots using the pyxem/orix library."""

# pylint: disable=no-member,duplicate-code,unsubscriptable-object,comparison-with-callable

# https://orix.readthedocs.io/en/stable/tutorials/inverse_pole_figures.html

# type: ignore

import os

from typing import Dict, Any, List

import matplotlib.pyplot as plt

import numpy as np

import h5py

# import imageio.v3 as iio
from PIL import Image as pil

import diffsims
import orix
from diffpy.structure import Lattice, Structure
from orix import plot
from orix.crystal_map import create_coordinate_arrays, CrystalMap, PhaseList
from orix.quaternion import Rotation
from orix.vector import Vector3d

from pynxtools.dataconverter.readers.em_om.utils.image_transform import thumbnail

from pynxtools.dataconverter.readers.em_om.utils.em_nexus_plots import HFIVE_WEB_MAX_SIZE

orix_params = {
    "figure.facecolor": "w",
    "figure.figsize": (8, 8),
    "lines.markersize": 10,
    "font.size": 15,
    "axes.grid": True,
}  # 20, 7),
plt.rcParams.update(orix_params)

# https://orix.readthedocs.io/ for details about the orix lib
# https://www.diffpy.org/diffpy.structure/package.html for details about the diffpy lib


class NxEmOmOrixEbsdParser:
    """Parse *.h5oina EBSD data.

    """

    def __init__(self, file_name, entry_id):
        """Class wrapping pyxem/orix H5OINA parser."""
        # this is one example which should be extended to kikuchipy but this should
        # be done together with Hakon and Knut Marthinsen's team or the pyxem developers
        # the reason why I went for a standalone parser was that kikuchipy v0.7.0 which
        # I had available when implementing this example was that *.h5oina files
        # were forced to have Kikuchi pattern in them. However, in *.h5oina the patterns
        # are optional and this was also the situation for the example file
        # which the EBSD system technology partner Oxford Instrument provided to me
        self.file_names = file_name
        self.entry_id = entry_id
        self.oina_version_identifier = "5.0"
        self.xaxis: List[float] = []
        self.yaxis: List[float] = []
        self.xmap = CrystalMap
        self.oina: Dict[str, Any] = {"n_slices": 1,
                                     "rotation": Rotation,
                                     "scan_point_x": [],
                                     "scan_point_y": [],
                                     "phase_identifier": [],
                                     "band_contrast": [],
                                     "scan_size": [0, 0],
                                     "scan_step": [0., 0.],
                                     "scan_unit": ["n/a", "n/a"],
                                     "phase": [],
                                     "space_group": []}
        # y (aka height), x (aka width) !

    def parse_h5oina(self, slice_id):
        """Parse content from *.h5oina format."""
        # https://github.com/oinanoanalysis/h5oina/blob/master/H5OINAFile.md
        # here implementing version 5.0
        # in this implementation we load all data in main memory, for stream processing
        # when the data are really large one would rather load the data array metadata
        # dtype, size, shape and use lazy loading instead or successive loading
        # file_names is file name
        h5r = h5py.File(self.file_names, "r")
        self.oina_version_identifier = h5r["/Format Version"][0].decode("utf-8")
        self.oina["n_slices"] = h5r["/Index"][0]
        print(f"H5OINA v{self.oina_version_identifier} has {self.oina['n_slices']} slices")
        if self.oina_version_identifier != "5.0" or self.oina["n_slices"] != 1:
            print("This examples supports H5OINA only in version 5.0 with one slice!")
            return

        self.parse_h5oina_ebsd_data(h5r, slice_id)
        self.parse_h5oina_ebsd_header(h5r, slice_id)

        group_names = []
        group_name = f"/{slice_id}/EBSD/Header/Phases"
        if group_name in h5r:
            group_names = sorted(list(h5r[group_name].keys()), key=int)
        for name in group_names:
            self.parse_h5oina_phase(h5r, slice_id, name)

        h5r.close()
        self.generate_xmap()

    def parse_h5oina_ebsd_data(self, h5r, slice_id):
        """Parse EBSD data section for specific slice."""
        group_name = f"/{slice_id}/EBSD/Data"
        # required entries in v5.0
        dset_name = f"{group_name}/Euler"
        if dset_name in h5r:
            self.oina["rotation"] = Rotation.from_euler(euler=h5r[dset_name],
                                                        direction='lab2crystal',
                                                        degrees=False)  # rad in v5.0
        dset_name = f"{group_name}/Phase"
        if dset_name in h5r:
            self.oina["phase_identifier"] = np.asarray(h5r[dset_name], np.int32)

        # optional entries in v5.0
        dset_name = f"{group_name}/X"  # v5.0, top-left-corner µm
        if dset_name in h5r:
            self.oina["scan_point_x"] = np.asarray(h5r[dset_name], np.float32)
        dset_name = f"{group_name}/Y"  # v5.0, top-left-corner µm
        if dset_name in h5r:
            self.oina["scan_point_y"] = np.asarray(h5r[dset_name], np.float32)
        dset_name = f"{group_name}/Band Contrast"
        if dset_name in h5r:  # is not required but should how else to create a ROI image
            self.oina["band_contrast"] = np.asarray(h5r[dset_name], np.uint8)

    def parse_h5oina_ebsd_header(self, h5r, slice_id):
        """Parse EBSD header section for specific slice."""
        group_name = f"/{slice_id}/EBSD/Header"
        dset_name = f"{group_name}/Y Cells"  # v5.0 height of the map
        if dset_name in h5r:
            self.oina["scan_size"][0] = h5r[dset_name][0]
        dset_name = f"{group_name}/X Cells"  # v5.0 width of the map
        if dset_name in h5r:
            self.oina["scan_size"][1] = h5r[dset_name][0]
        dset_name = f"{group_name}/Y Step"  # v5.0, µm
        if dset_name in h5r:
            dst_y = h5r[dset_name]
            self.oina["scan_step"][0] = dst_y[0]
            if "Unit" in dst_y.attrs:
                self.oina["scan_unit"][0] = dst_y.attrs["Unit"]
        dset_name = f"{group_name}/X Step"  # v5.0, µm
        if dset_name in h5r:
            dst_x = h5r[dset_name]
            self.oina["scan_step"][1] = dst_x[0]
            if "Unit" in dst_x.attrs:
                self.oina["scan_unit"][1] = dst_x.attrs["Unit"]
        # ##MK:: check that all data in the self.oina are consistent

    def parse_h5oina_phase(self, h5r, slice_id, name):
        """Parse one phase."""
        print(f"Loading phase {name}...")
        # required in v5.0
        sub_group_name = f"/{slice_id}/EBSD/Header/Phases/{name}"
        dset_name = f"{sub_group_name}/Phase Name"
        if dset_name in h5r:
            phase_name = h5r[dset_name][0].decode("utf-8")
        # "Reference", but even examples from Oxford place no DOIs here
        dset_name = f"{sub_group_name}/Lattice Angles"
        if dset_name in h5r:
            alpha_beta_gamma = np.asarray(h5r[dset_name][:].flatten()) / np.pi * 180.
            # rad2deg
        dset_name = f"{sub_group_name}/Lattice Dimensions"
        if dset_name in h5r:
            a_b_c = np.asarray(h5r[dset_name][:].flatten()) * 0.1
            # angstroem2nm

        # optional in v5.0
        # many inconsistencies and metadata cluttering by design across
        # tools of the community and of technology partners:
        # e.g. MTex uses point group required, Oxford uses space group but
        # optional, neither of them store atom positions and
        # computed intensities because for this there are other files...
        dset_name = f"{sub_group_name}/Space Group"
        if dset_name in h5r:
            space_group = int(h5r[dset_name][0])
            self.oina["space_group"].append(space_group)

        self.oina["phase"].append(
            Structure(title=phase_name,
                      atoms=None,
                      lattice=Lattice(a_b_c[0], a_b_c[1], a_b_c[2],
                                      alpha_beta_gamma[0],
                                      alpha_beta_gamma[1],
                                      alpha_beta_gamma[2])))

    def generate_xmap(self):
        """Generate the orientation map and orix/diffsims data structures."""
        coordinates, _ = create_coordinate_arrays(
            (self.oina["scan_size"][0], self.oina["scan_size"][1]),
            (self.oina["scan_step"][0], self.oina["scan_step"][1]))
        self.xaxis = coordinates["x"]
        self.yaxis = coordinates["y"]
        del coordinates

        # orix expects non-indexed scan points to have phase_id -1
        # while H5OINA v5.0 specifies that such points have phase ID 0
        # H5OINA v5.0 specifies further one subgroup per phase
        # phase_identifier starting at 1, as non-indexed points do not belong to a phase
        self.oina["phase_identifier"] = self.oina["phase_identifier"] - 1
        print(np.unique(self.oina["phase_identifier"]))

        self.xmap = CrystalMap(rotations=self.oina["rotation"],
                               x=self.xaxis, y=self.yaxis,
                               phase_id=self.oina["phase_identifier"],
                               phase_list=PhaseList(space_groups=self.oina["space_group"],
                                                    structures=self.oina["phase"]),
                               prop={"bc": self.oina["band_contrast"]},
                               scan_unit=self.oina["scan_unit"])
        print(self.xmap)

    def parse_roi_default_plot(self, template: dict) -> dict:
        """Parse data for the region-of-interest default plot."""
        print("Parse ROI default plot...")
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/region_of_interest"
        template[f"{trg}/descriptor"] = str("normalized_band_contrast")
        # ##MK:: need a better strategy here because the descriptor
        # band contrast is not a required field in v5.0,
        # thus it should not be hardcoded and assumed that it is present
        # but there is also no other field in v5.0 that could stand in as a descriptor

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/region_of_interest/roi"
        template[f"{trg}/title"] = str("Region-of-interest overview image")
        template[f"{trg}/@signal"] = "data"
        template[f"{trg}/@axes"] = ["axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/region_of_interest/roi/data"
        template[f"{trg}"] = {"compress": np.reshape(
            np.asarray(np.asarray((self.xmap.bc / np.max(self.xmap.bc) * 255.),
                       np.uint32), np.uint8), (self.xmap.shape[0], self.xmap.shape[1]),
            order="C"), "strength": 1}
        # 0 is y while 1 is x !
        template[f"{trg}/@long_name"] = "Signal"
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/region_of_interest/roi/axis_x"
        template[f"{trg}"] = {"compress": np.asarray(
            0.5 * self.oina["scan_step"][1] + self.xaxis[0:self.xmap.shape[1]],
            np.float32), "strength": 1}
        template[f"{trg}/@long_name"] \
            = f"Calibrated coordinate along x-axis ({self.oina['scan_unit'][1]})"
        template[f"{trg}/@units"] = self.oina["scan_unit"][1]

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/region_of_interest/roi/axis_y"
        template[f"{trg}"] = {"compress": np.asarray(
            0.5 * self.oina["scan_step"][0] + self.yaxis[0:self.xmap.size:self.xmap.shape[1]],
            np.float32), "strength": 1}
        template[f"{trg}/@long_name"] \
            = f"Calibrated coordinate along y-axis ({self.oina['scan_unit'][0]})"
        template[f"{trg}/@units"] = self.oina["scan_unit"][0]
        return template

    def parse_phases(self, template: dict) -> dict:
        """Parse data for the crystal structure models aka phases."""
        print("Parse crystal_structure_models aka phases...")
        identifier = 1
        # identifier match because phase is a list asc. sorted by numerical keys
        for phase in self.oina["phase"]:
            trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
                  f"EM_EBSD_CRYSTAL_STRUCTURE_MODEL" \
                  f"[em_ebsd_crystal_structure_model{identifier}]"
            template[f"{trg}/phase_identifier"] = np.uint32(identifier)
            template[f"{trg}/phase_name"] = str(phase.title)
            template[f"{trg}/unit_cell_abc"] = np.asarray(
                [phase.lattice.a, phase.lattice.b, phase.lattice.c],
                np.float32)
            template[f"{trg}/unit_cell_abc/@units"] = "nm"
            template[f"{trg}/unit_cell_alphabetagamma"] = np.asarray(
                [phase.lattice.alpha, phase.lattice.beta, phase.lattice.gamma],
                np.float32)
            template[f"{trg}/unit_cell_alphabetagamma/@units"] = "°"
            identifier += 1
        return template

    def parse_inverse_pole_figures(self, template: dict) -> dict:
        """Parse inverse pole figures (IPF) mappings."""
        print("Parse inverse pole figures (IPFs)...")
        print(len(self.oina["phase"]))
        # identifier match because phase is a list asc. sorted by numerical keys
        for identifier in np.arange(0, np.max(self.xmap.phase_id) + 1):
            self.parse_inverse_pole_figure_map(identifier, template)
            self.parse_inverse_pole_figure_color_key(identifier, template)
            identifier += 1

        return template

    def parse_inverse_pole_figure_map(self, identifier, template: dict) -> dict:
        """Parse and create inverse-pole-figure (IPF) mappings on their color models."""
        # +1 because for orix not_indexed -1 and "first" phase has ID 0 !
        phase_id = identifier + 1
        phase_name = self.xmap.phases[identifier].name
        print(f"Generate inverse pole figure (IPF) map for {identifier}, {phase_name}...")

        phase_id_ipf_key = plot.IPFColorKeyTSL(
            self.xmap.phases[identifier].point_group.laue,
            direction=Vector3d.zvector())

        rgb_px_with_phase_id = np.asarray(
            np.asarray(phase_id_ipf_key.orientation2color(
                self.xmap[phase_name].rotations) * 255., np.uint32), np.uint8)

        ipf_rgb_map = np.asarray(
            np.uint8(np.zeros((self.xmap.shape[0] * self.xmap.shape[1], 3)) * 255.0))
        # background is black instead of white (which would be more pleasing)
        # but IPF color maps have a whitepoint which encodes in fact an orientation
        # and because of that we may have a single crystal with an orientation
        # close to the whitepoint which become a fully white seemingly "empty" image
        ipf_rgb_map[self.xmap.phase_id == identifier, :] = rgb_px_with_phase_id
        ipf_rgb_map = np.reshape(ipf_rgb_map,
                                 (self.xmap.shape[0], self.xmap.shape[1], 3),
                                 order="C")  # 0 is y while 1 is x !

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{phase_id}]"
        template[f"{trg}/bitdepth"] = np.uint32(8)
        template[f"{trg}/phase_identifier"] = np.uint32(phase_id)
        template[f"{trg}/phase_name"] = str(phase_name)
        template[f"{trg}/PROGRAM[program1]/program"] = str("orix")
        template[f"{trg}/PROGRAM[program1]/program/@version"] = orix.__version__
        template[f"{trg}/PROGRAM[program2]/program"] = str("diffsims")
        template[f"{trg}/PROGRAM[program2]/program/@version"] = diffsims.__version__
        template[f"{trg}/projection_direction"] = np.asarray([0., 0., 1.], np.float32)
        # should have a reference so that it is directly interpretable

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{phase_id}]/ipf_rgb_map"
        template[f"{trg}/title"] = str("Inverse pole figure color map")
        template[f"{trg}/@signal"] = "data"
        template[f"{trg}/@axes"] = ["axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{phase_id}]/ipf_rgb_map/DATA[data]"
        template[f"{trg}"] = {"compress": ipf_rgb_map, "strength": 1}
        template[f"{trg}/@CLASS"] = "IMAGE"  # required by H5Web to plot RGB maps
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        # dimension scale axes value arrays same for each phase, entire IPF map
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{phase_id}]/ipf_rgb_map/axis_x"
        template[f"{trg}"] = {"compress": np.asarray(
            0.5 * self.oina["scan_step"][1] + self.xaxis[0:self.xmap.shape[1]],
            np.float32), "strength": 1}
        template[f"{trg}/@long_name"] \
            = f"Calibrated coordinate along x-axis ({self.oina['scan_unit'][1]})"
        template[f"{trg}/@units"] = self.oina["scan_unit"][1]

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/" \
              f"PROCESS[ipf_map{phase_id}]/ipf_rgb_map/axis_y"
        template[f"{trg}"] = {"compress": np.asarray(
            0.5 * self.oina["scan_step"][0]
            + self.yaxis[0:self.xmap.size:self.xmap.shape[1]],
            np.float32), "strength": 1}
        template[f"{trg}/@long_name"] \
            = f"Calibrated coordinate along y-axis ({self.oina['scan_unit'][0]})"
        template[f"{trg}/@units"] = self.oina["scan_unit"][0]

        return template

    def parse_inverse_pole_figure_color_key(self, identifier, template: dict) -> dict:
        """Parse color key renderings of inverse-pole-figure (IPF) mappings."""
        # +1 because for orix not_indexed -1 and "first" phase has ID 0 !
        phase_id = identifier + 1
        phase_name = self.xmap.phases[identifier].name
        print(f"Parse inverse pole figure (IPF) color key {identifier}, {phase_name}...")

        phase_id_ipf_key = plot.IPFColorKeyTSL(
            self.xmap.phases[identifier].point_group.laue,
            direction=Vector3d.zvector())

        # render domain-specific IPF color keys using orix
        fig = phase_id_ipf_key.plot(return_figure=True)
        fig.savefig("temporary.png", dpi=300, facecolor='w', edgecolor='w',
                    orientation='landscape', format='png', transparent=False,
                    bbox_inches='tight', pad_inches=0.1, metadata=None)
        # constraint further to 8bit RGB and no flipping
        # im = np.asarray(imageio.v3.imread(symm_name))
        img = np.asarray(thumbnail(pil.open("temporary.png", "r", ["png"]),
                         size=HFIVE_WEB_MAX_SIZE), np.uint8)
        img = img[:, :, 0:3]  # discard alpha channel
        if os.path.exists("temporary.png"):
            os.remove("temporary.png")
        # ##MK::need to constrain more the writing of the image that it is guaranteed
        # a specific type of image and bitdepth and color model, and avoid implicit
        # image transformations such as flips or rotations

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model"
        template[f"{trg}/title"] = str("Inverse pole figure color key with SST")
        template[f"{trg}/@signal"] = "data"
        template[f"{trg}/@axes"] = ["axis_y", "axis_x"]
        template[f"{trg}/@AXISNAME_indices[axis_x_indices]"] = np.uint32(0)
        template[f"{trg}/@AXISNAME_indices[axis_y_indices]"] = np.uint32(1)
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model/DATA[data]"
        template[f"{trg}"] = {"compress": img, "strength": 1}
        template[f"{trg}/@CLASS"] = "IMAGE"
        template[f"{trg}/@IMAGE_VERSION"] = "1.2"
        template[f"{trg}/@SUBCLASS_VERSION"] = np.int64(15)

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model/AXISNAME[axis_y]"
        template[f"{trg}"] = {"compress":
                              np.asarray(np.linspace(1,
                                                     np.shape(img)[0],
                                                     num=np.shape(img)[0],
                                                     endpoint=True), np.uint32),
                              "strength": 1}
        template[f"{trg}/@long_name"] = "Pixel along y-axis"
        template[f"{trg}/@units"] = "px"

        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing" \
              f"/PROCESS[ipf_map{phase_id}]/ipf_rgb_color_model/AXISNAME[axis_x]"
        template[f"{trg}"] = {"compress":
                              np.asarray(np.linspace(1,
                                                     np.shape(img)[1],
                                                     num=np.shape(img)[1],
                                                     endpoint=True), np.uint32),
                              "strength": 1}
        template[f"{trg}/@long_name"] = "Pixel along x-axis"
        template[f"{trg}/@units"] = "px"

        return template

    def parse(self, template: dict) -> dict:
        """Parse NOMAD OASIS relevant data and metadata from an H5OINA file."""
        print("Parsing EBSD data pyxem/orix-style for an H5OINA example...")
        print(self.file_names)
        print(f"{self.entry_id}")
        self.parse_h5oina(1)
        self.parse_roi_default_plot(template)
        self.parse_phases(template)
        self.parse_inverse_pole_figures(template)
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Create nionswift mapping from ns UUID to suffix used for identifying files in project."""

# pylint: disable=no-member,too-many-instance-attributes

import uuid

# see https://github.com/nion-software/nionswift/blob/e95839c5602d009006ea88a648e5f78dc77c1ea4/
# nion/swift/model/Profile.py line 146 and following


def encode(uuid_: uuid.UUID, alphabet: str) -> str:
    result = str()
    uuid_int = uuid_.int
    while uuid_int:
        uuid_int, digit = divmod(uuid_int, len(alphabet))
        result += alphabet[digit]
    return result


def uuid_to_file_name(data_item_uuid_str: str) -> str:
    data_item_uuid_uuid = uuid.UUID(f'{data_item_uuid_str}')
    return f'data_{encode(data_item_uuid_uuid, "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890")}'
    # 25 character results



import argparse
import sys

from .apps import dir_app
from .apps import impatient_app
from .apps import manual_app
from .apps import nxclass_app


def main(argv=None):
    parser = argparse.ArgumentParser(description="NeXus development tools")

    subparsers = parser.add_subparsers(help="Commands", dest="command")

    nxclass_parser = subparsers.add_parser(
        "nxclass", help="Test and documentation for a single NeXus class"
    )
    nxclass_app.nxclass_args(nxclass_parser)
    dir_app.dir_args(nxclass_parser)

    manual_parser = subparsers.add_parser(
        "manual", help="Test and prepare User Manual building"
    )
    manual_app.manual_args(manual_parser)
    dir_app.dir_args(manual_parser)

    impatient_parser = subparsers.add_parser(
        "impatient", help="Prepare Impatient Guide building"
    )
    impatient_app.impatient_args(impatient_parser)
    dir_app.dir_args(impatient_parser)

    if argv is None:
        argv = sys.argv
    args = parser.parse_args(argv[1:])

    if args.command == "nxclass":
        dir_app.dir_exec(args)
        nxclass_app.nxclass_exec(args)
    elif args.command == "manual":
        dir_app.dir_exec(args)
        manual_app.manual_exec(args)
    elif args.command == "impatient":
        dir_app.dir_exec(args)
        impatient_app.impatient_exec(args)
    else:
        parser.print_help()
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""This module intended to generate schema eln which usually randeredto NOMAD."""

# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Dict, Any
import xml.etree.ElementTree as ET
import yaml
from pynxtools.eln_mapper.eln import retrieve_nxdl_file
from pynxtools.dataconverter.helpers import remove_namespace_from_tag


NEXUS_TYPE_TO_NUMPY_TYPE = {'NX_CHAR': {'convert_typ': 'str',
                                        'component_nm': 'StringEditQuantity',
                                        'default_unit_display': '<No Default unit>'},
                            'NX_BOOLEAN': {'convert_typ': 'bool',
                                           'component_nm': 'BoolEditQuantity',
                                           'default_unit_display': '<No Default unit>'},
                            'NX_DATE_TIME': {'convert_typ': 'Datetime',
                                             'component_nm': 'DateTimeEditQuantity',
                                             'default_unit_display': '<No Default unit>'},
                            'NX_FLOAT': {'convert_typ': 'np.float64',
                                         'component_nm': 'NumberEditQuantity',
                                         'default_unit_display': '<No Default unit>'},
                            'NX_INT': {'convert_typ': 'int',
                                       'component_nm': 'NumberEditQuantity',
                                       'default_unit_display': '<No Default unit>'},
                            'NX_NUMBER': {'convert_typ': 'np.float64',
                                          'component_nm': 'NumberEditQuantity',
                                          'default_unit_display': '<No Default unit>'},
                            '<NO FILED TYPE>': {'convert_typ': '<NO FILED TYPE>',
                                                'component_nm': '<No Edit Quantity>',
                                                'default_unit_display': '<No Default unit>'},
                            }


def construct_field_structure(fld_elem, quntities_dict):
    """Construct field structure such as unit, value.
    Parameters
    ----------
    elem : _type_
        _description_
    quntities_dict : _type_
        _description_
    """
    elm_attr = fld_elem.attrib
    fld_nm = elm_attr['name'].lower()
    quntities_dict[fld_nm] = {}
    fld_dict = quntities_dict[fld_nm]

    # handle type
    if 'type' in elm_attr:
        nx_fld_typ = elm_attr['type']
    else:
        nx_fld_typ = 'NX_CHAR'

    if nx_fld_typ in NEXUS_TYPE_TO_NUMPY_TYPE:
        cov_fld_typ = NEXUS_TYPE_TO_NUMPY_TYPE[nx_fld_typ]['convert_typ']

    fld_dict['type'] = cov_fld_typ
    if 'units' in elm_attr:
        fld_dict['unit'] = f"<hint: {elm_attr['units']}>"
        fld_dict['value'] = "<ADD default value>"

    # handle m_annotation
    m_annotation = {'m_annotations': {'eln':
                                      {'component':
                                       NEXUS_TYPE_TO_NUMPY_TYPE[nx_fld_typ]['component_nm'],
                                       'defaultDisplayUnit':
                                       (NEXUS_TYPE_TO_NUMPY_TYPE[nx_fld_typ]
                                        ['default_unit_display'])}}}
    fld_dict.update(m_annotation)

    # handle description
    construct_decription(fld_elem, fld_dict)


def construct_decription(elm: ET.Element, concept_dict: Dict) -> None:
    """Collect doc from concept doc.
    """
    desc_text = ''
    for child_elm in elm:
        tag = remove_namespace_from_tag(child_elm.tag)
        if tag == 'doc':
            desc_text = child_elm.text
            desc_text = ' '.join([x.strip() for x in desc_text.split('\n')])
            break

    concept_dict['description'] = desc_text


def construct_group_structure(grp_elm: ET.Element, subsections: Dict) -> None:
    """To construct group structure as follows:
    <group_name>:
        section:
            m_annotations:
                eln:
                    overview: true

    Parameters
    ----------
    elm : ET.Element
        Group element
    subsections : Dict
        Dict to include group recursively
    """

    default_m_annot = {'m_annotations': {'eln': {'overview': True}}}

    elm_attrib = grp_elm.attrib
    grp_desig = ""
    if 'name' in elm_attrib:
        grp_desig = elm_attrib['name'].capitalize()
    elif 'type' in elm_attrib:
        grp_desig = elm_attrib['type'][2:].capitalize()

    subsections[grp_desig] = {}
    grp_dict = subsections[grp_desig]

    # add setion in group
    grp_dict['section'] = {}
    section = grp_dict['section']
    section.update(default_m_annot)

    # pass the grp elment for recursive search
    scan_xml_element_recursively(grp_elm, section)


def _should_skip_iteration(elm: ET.Element) -> bool:
    """Define some elements here that should be skipped.

    Parameters
    ----------
    elm : ET.Element
        The element to investigate to skip
    """
    attr = elm.attrib
    elm_type = ''
    if 'type' in attr:
        elm_type = attr['type']
        if elm_type in ['NXentry']:
            return True
    return False


def scan_xml_element_recursively(nxdl_element: ET.Element,
                                 recursive_dict: Dict,
                                 root_name: str = "",
                                 reader_name: str = '<READER_NAME>',
                                 is_root: bool = False) -> None:
    """Scan xml elements, and pass the element to the type of element handaler.

    Parameters
    ----------
    nxdl_element : ET.Element
        This xml element that will be scanned through the descendants.
    recursive_dict : Dict
        A dict that store hierarchical structure of scheme eln.
    root_name : str, optional
        Name of root that user want to see to name their application, e.g. MPES,
        by default 'ROOT_NAME'
    reader_name : Prefered name of the reader.
    is_root : bool, optional
        Declar the elment as root or not, by default False
    """

    if is_root:
        # Note for later: crate a new function to handle root part
        nxdl = 'NX<NAME>.nxdl'
        recursive_dict[root_name] = {'base_sections':
                                     ['nomad.datamodel.metainfo.eln.NexusDataConverter',
                                      'nomad.datamodel.data.EntryData']}

        m_annotations: Dict = {'m_annotations': {'template': {'reader': reader_name,
                                                              'nxdl': nxdl},
                                                 'eln': {'hide': []}}}

        recursive_dict[root_name].update(m_annotations)

        recursive_dict = recursive_dict[root_name]

    # Define quantities for taking care of field
    quantities: Dict = None
    subsections: Dict = None
    for elm in nxdl_element:
        tag = remove_namespace_from_tag(elm.tag)
        # To skip NXentry group but only consider the child elments
        if _should_skip_iteration(elm):
            scan_xml_element_recursively(elm, recursive_dict)
            continue
        if tag == 'field':
            if quantities is None:
                recursive_dict['quantities'] = {}
                quantities = recursive_dict['quantities']
            construct_field_structure(elm, quantities)
        if tag == 'group':
            if subsections is None:
                recursive_dict['sub_sections'] = {}
                subsections = recursive_dict['sub_sections']
            construct_group_structure(elm, subsections)


def get_eln_recursive_dict(recursive_dict: Dict, nexus_full_file: str) -> None:
    """Develop a recursive dict that has hierarchical structure of scheme eln.

    Parameters
    ----------
    recursive_dict : Dict
        A dict that store hierarchical structure of scheme eln.
    nexus_full_file : str
        Full path of NeXus file e.g. <full_path>/paNXmpes.nxdl.xml
    """

    nxdl_root = ET.parse(nexus_full_file).getroot()
    root_name = nxdl_root.attrib['name'][2:] if 'name' in nxdl_root.attrib else "<ROOT_NAME>"
    recursive_dict['definitions'] = {'name': '<ADD PREFERED NAME>',
                                     'sections': {}}
    sections = recursive_dict['definitions']['sections']

    scan_xml_element_recursively(nxdl_root, sections,
                                 root_name=root_name, is_root=True)


def generate_scheme_eln(nexus_def: str, eln_file_name: str = None) -> None:
    """Generate schema eln that should go to Nomad while running the reader.
    The output file will be <eln_file_name>.scheme.archive.yaml

    Parameters
    ----------
    nexus_def : str
        Name of nexus definition e.g. NXmpes
    eln_file_name : str
        Name of output file e.g. mpes

    Returns:
        None
    """

    file_parts: list = []
    out_file_ext = 'scheme.archive.yaml'
    raw_name = ""
    out_file = ""

    nxdl_file = retrieve_nxdl_file(nexus_def)

    if eln_file_name is None:
        # raw_name from e.g. /<path>/NXmpes.nxdl.xml
        raw_name = nxdl_file.split('/')[-1].split('.')[0][2:]
        out_file = '.'.join([raw_name, out_file_ext])
    else:
        file_parts = eln_file_name.split('.')
        if len(file_parts) == 1:
            raw_name = file_parts[0]
            out_file = '.'.join([raw_name, out_file_ext])
        elif len(file_parts) == 4 and '.'.join(file_parts[1:]) == out_file_ext:
            out_file = eln_file_name
        elif nexus_def[0:2] == 'NX':
            raw_name = nexus_def[2:]
            out_file = '.'.join([raw_name, out_file_ext])
        else:
            raise ValueError("Check for correct NeXus definition and output file name.")

    recursive_dict: Dict[str, Any] = {}
    get_eln_recursive_dict(recursive_dict, nxdl_file)
    # print('recursive_dict', recursive_dict)

    with open(out_file, mode='w', encoding='utf-8') as out_f:
        yaml.dump(recursive_dict, sort_keys=False, stream=out_f)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility class handling which community/technology partner files for em_nion reader."""

# pylint: disable=no-member,duplicate-code,too-few-public-methods

from typing import Tuple, Dict, List


class EmNionUseCaseSelector:
    """Decision maker about what needs to be parsed given arbitrary input.

    Users might invoke this dataconverter with arbitrary input, no input, or
    too much input. The UseCaseSelector decides what to do in each case.
    """

    def __init__(self, file_paths: Tuple[str] = None):
        """Analyze with which parser(s) (if any) input can be handled."""
        self.mime_types: Dict[str, list] = {}
        self.eln_parser_type: str = "none"
        self.eln: List[str] = []
        self.prj_parser_type: str = "none"
        self.prj: List[str] = []
        self.is_valid = False

        self.analyze_mime_types(file_paths)
        self.identify_case()

    def analyze_mime_types(self, file_paths: Tuple[str] = None):
        """Accept/reject and organize input-files based on mime-type."""
        self.supported_mime_types = ["yaml", "yml", "nionswift"]
        for mime_type in self.supported_mime_types:
            self.mime_types[mime_type] = []
        for file_name in file_paths:
            index = file_name.lower().rfind(".")
            if index >= 0:
                suffix = file_name.lower()[index + 1::]
                add = (suffix in self.supported_mime_types) \
                    and (file_name not in self.mime_types[suffix])
                if add is True:
                    self.mime_types[suffix].append(file_name)
        print(self.mime_types)

    def identify_case(self):
        """Identify which sub-parsers to use if any based on input mime_types."""
        # eln
        self.is_valid = False
        for mime_type in ["yaml", "yml"]:
            self.eln += self.mime_types[mime_type]
        if len(self.eln) == 1:
            self.eln_parser_type = "generic"
            self.is_valid = True
        else:
            self.eln_parser_type = "none"

        if "nionswift" in self.mime_types:
            if len(self.mime_types["nionswift"]) == 1:
                self.prj += self.mime_types["nionswift"]
                self.prj_parser_type = "nionswift"
            else:
                self.is_valid = False

        print("Input suggests to use the following sub-parsers:")
        print(f"ELN parser: {self.eln_parser_type}")
        print(self.eln)
        print(f"Data parser: {self.prj_parser_type}")
        print(self.prj)
        print(f"Input suggests that parsing is valid: {self.is_valid}")



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test cases for the Writer class used by the DataConverter"""

import os

import pytest
import h5py
from pynxtools.dataconverter.exceptions import InvalidDictProvided


from pynxtools.dataconverter.writer import Writer
from .test_helpers import fixture_filled_test_data, fixture_template, alter_dict  # pylint: disable=unused-import


@pytest.mark.usefixtures("filled_test_data")
@pytest.fixture(name="writer")
def fixture_writer(filled_test_data, tmp_path):
    """pytest fixture to setup Writer object to be used by tests with dummy data."""
    writer = Writer(
        filled_test_data,
        os.path.join("tests", "data", "dataconverter", "NXtest.nxdl.xml"),
        os.path.join(tmp_path, "test.nxs")
    )
    yield writer
    del writer


def test_init(writer):
    """Test to verify Writer's initialization works."""
    assert isinstance(writer, Writer)


def test_write(writer):
    """Test for the Writer's write function. Checks whether entries given above get written out."""
    writer.write()
    test_nxs = h5py.File(writer.output_path, "r")
    assert test_nxs["/my_entry/NXODD_name/int_value"][()] == 2
    assert test_nxs["/my_entry/NXODD_name/int_value"].attrs["units"] == "eV"
    assert test_nxs["/my_entry/NXODD_name/posint_value"].shape == (3,)  # pylint: disable=no-member


def test_write_link(writer):
    """Test for the Writer's write function.

Checks whether entries given above get written out when a dictionary containing a link is
given in the template dictionary."""
    writer.write()
    test_nxs = h5py.File(writer.output_path, "r")
    assert isinstance(test_nxs["/my_entry/links/ext_link"], h5py.Dataset)


@pytest.mark.usefixtures("filled_test_data")
def test_wrong_dict_provided_in_template(filled_test_data, tmp_path):
    """Tests if the writer correctly fails when a wrong dictionary is provided"""
    writer = Writer(
        alter_dict(filled_test_data,
                   "/ENTRY[my_entry]/links/ext_link",
                   {"not a link or anything": 2.0}),
        os.path.join("tests", "data", "dataconverter", "NXtest.nxdl.xml"),
        os.path.join(tmp_path, "test.nxs")
    )
    with pytest.raises(InvalidDictProvided) as execinfo:
        writer.write()
        assert str(execinfo.value) == ("pynxtools.dataconverter.exceptions.InvalidDictProvided: "
                                       "A dictionary was provided to the template but it didn't "
                                       "fall into any of the know cases of handling dictionaries"
                                       ". This occured for: ext_link")



#!/usr/bin/env python
"""Reads NeXus HDF5 files using h5py and prints the contents"""

import h5py  # HDF5 support

fileName = "simple_example_basic.nexus.hdf5"
with h5py.File(fileName, "r") as f:
    for item in f.attrs.keys():
        print(item + ":", f.attrs[item])
    mr = f["/entry/mr_scan/mr"]
    i00 = f["/entry/mr_scan/I00"]
    print("%s\t%s\t%s" % ("#", "mr", "I00"))
    for i in range(len(mr)):
        print("%d\t%g\t%d" % (i, mr[i], i00[i]))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility tool constants and versioning."""

# pylint: disable=no-member

from pynxtools.dataconverter.readers.shared.shared_utils \
    import get_repo_last_commit


NX_EM_NION_ADEF_NAME = "NXem"
NX_EM_NION_ADEF_VERSION = "nexus-fairmat-proposal successor of " \
                          "9636feecb79bb32b828b1a9804269573256d7696"
# based on https://fairmat-experimental.github.io/nexus-fairmat-proposal
NX_EM_NION_EXEC_NAME = "dataconverter/reader/em_nion/reader.py"
NX_EM_NION_EXEC_VERSION = get_repo_last_commit()

NX_EM_NION_SWIFT_NAME = "nionswift"
NX_EM_NION_SWIFT_VERSION = "0.16.8"



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test cases for readers used for the DataConverter"""

import glob
import os
from typing import List
import xml.etree.ElementTree as ET

import pytest
from _pytest.mark.structures import ParameterSet

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.convert import \
    get_names_of_all_readers, get_reader
from pynxtools.dataconverter.helpers import \
    validate_data_dict, generate_template_from_nxdl
from pynxtools.dataconverter.template import Template


def get_reader_name_from_reader_object(reader) -> str:
    """Helper function to find the name of a reader given the reader object."""
    for reader_name in get_names_of_all_readers():
        gotten_reader = get_reader(reader_name)
        if reader.__name__ is gotten_reader.__name__:
            return reader_name
    return ""


def get_readers_file_names() -> List[str]:
    """Helper function to parametrize paths of all the reader Python files"""
    return sorted(glob.glob("pynxtools/dataconverter/readers/*/reader.py"))


def get_all_readers() -> List[ParameterSet]:
    """Scans through the reader list and returns them for pytest parametrization"""
    readers = []

    # Explicitly removing ApmReader and EmNionReader because we need to add test data
    for reader in [get_reader(x) for x in get_names_of_all_readers()]:
        if reader.__name__ in ("ApmReader", "EmOmReader", "EmSpctrscpyReader", "EmNionReader"):
            readers.append(pytest.param(reader,
                                        marks=pytest.mark.skip(reason="Missing test data.")
                                        ))
        else:
            readers.append(pytest.param(reader))

    return readers


@pytest.mark.parametrize("reader", get_all_readers())
def test_if_readers_are_children_of_base_reader(reader):
    """Test to verify that all readers are children of BaseReader"""
    if reader.__name__ != "BaseReader":
        assert isinstance(reader(), BaseReader)


@pytest.mark.parametrize("reader", get_all_readers())
def test_has_correct_read_func(reader):
    """Test if all readers have a valid read function implemented"""
    assert callable(reader.read)
    if reader.__name__ not in ["BaseReader"]:
        assert hasattr(reader, "supported_nxdls")

        reader_name = get_reader_name_from_reader_object(reader)
        def_dir = os.path.join(os.getcwd(), "pynxtools", "definitions")
        dataconverter_data_dir = os.path.join("tests", "data", "dataconverter")

        input_files = sorted(
            glob.glob(os.path.join(dataconverter_data_dir, "readers", reader_name, "*"))
        )
        for supported_nxdl in reader.supported_nxdls:
            if supported_nxdl in ("NXtest", "*"):
                nxdl_file = os.path.join(dataconverter_data_dir, "NXtest.nxdl.xml")
            elif supported_nxdl == "NXroot":
                nxdl_file = os.path.join(def_dir, "base_classes", "NXroot.nxdl.xml")
            else:
                nxdl_file = os.path.join(
                    def_dir, "contributed_definitions", f"{supported_nxdl}.nxdl.xml"
                )

            root = ET.parse(nxdl_file).getroot()
            template = Template()
            generate_template_from_nxdl(root, template)

            read_data = reader().read(template=Template(template), file_paths=tuple(input_files))

            assert isinstance(read_data, Template)
            assert validate_data_dict(template, read_data, root)


@pytest.mark.parametrize("reader_name,nxdl,undocumented_keys", [
    ('mpes', 'NXmpes', [])
])
def test_shows_correct_warnings(reader_name, nxdl, undocumented_keys):
    """
    Checks whether the read function generates the correct warnings.
    """
    def_dir = os.path.join(os.getcwd(), "pynxtools", "definitions")
    dataconverter_data_dir = os.path.join("tests", "data", "dataconverter")

    input_files = sorted(
        glob.glob(os.path.join(dataconverter_data_dir, "readers", reader_name, "*"))
    )
    nxdl_file = os.path.join(
        def_dir, "contributed_definitions", f"{nxdl}.nxdl.xml"
    )

    root = ET.parse(nxdl_file).getroot()
    template = Template()
    generate_template_from_nxdl(root, template)

    read_data = get_reader(reader_name)().read(
        template=Template(template), file_paths=tuple(input_files)
    )

    assert validate_data_dict(template, read_data, root)
    assert list(read_data.undocumented.keys()) == undocumented_keys



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""The writer class for writing a NeXus file in accordance with a given NXDL."""

# pylint: disable=R0912

import copy
import logging
import sys
import xml.etree.ElementTree as ET

import h5py
import numpy as np

from pynxtools.dataconverter import helpers
from pynxtools.nexus import nexus
from pynxtools.dataconverter.exceptions import InvalidDictProvided

logger = logging.getLogger(__name__)  # pylint: disable=C0103
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def does_path_exist(path, h5py_obj) -> bool:
    """Returns true if the requested path exists in the given h5py object."""
    try:
        h5py_obj[helpers.convert_data_dict_path_to_hdf5_path(path)]  # pylint: disable=W0106
        return True
    except KeyError:
        return False


def is_not_data_empty(value) -> bool:
    """Returns True if value is an Numpy array or not None."""
    if isinstance(value, np.ndarray) or value is not None:
        return True
    return False


def get_namespace(element) -> str:
    """Extracts the namespace for elements in the NXDL"""
    return element.tag[element.tag.index("{"):element.tag.rindex("}") + 1]


def split_link(data, output_path):
    """Handle the special syntax used in the reader for the dataset links.

Split the file:path variable in two variables file and path.
If multiple datasets are provided, the function returns two lists"""

    if not isinstance(data['link'], list):
        if ':' in data['link']:
            file = data['link'].split(":", 1)[0]
            path = data['link'].split(":", 1)[1]
        elif ':' not in data['link']:
            file = output_path
            path = data['link']
    else:
        file = []
        path = []
        for dataset in data['link']:
            if ':' in dataset:
                file.append(dataset.split(":", 1)[0])
                path.append(dataset.split(":", 1)[1])
            elif ':' not in data['link']:
                file.append(output_path)
                path.append(dataset)

    return file, path


def handle_shape_entries(data, file, path):
    """slice generation via the key shape"""
    new_shape = []
    for dim, val in enumerate(data['shape']):
        if isinstance(val, slice):
            start = val.start if val.start is not None else 0
            stop = val.stop if val.stop is not None else h5py.File(file, 'r')[path].shape[dim]
            step = val.step if val.step is not None else 1
            new_shape.append(int((stop - start) / step))
    if not new_shape:
        new_shape = [1]
    layout = h5py.VirtualLayout(shape=tuple(new_shape),
                                dtype=np.float64)
    vsource = h5py.VirtualSource(file,
                                 path,
                                 shape=h5py.File(file, 'r')[path].shape
                                 )[data['shape']]
    layout[:] = vsource
    return layout


# pylint: disable=too-many-locals, inconsistent-return-statements
def handle_dicts_entries(data, grp, entry_name, output_path, path):
    """Handle function for dictionaries found as value of the nexus file.

Several cases can be encoutered:
- Data to slice and place in virtual datasets
- Concatenate dataset in one virtual dataset
- Internal links
- External links
- compression label
"""
    if 'link' in data:
        file, path = split_link(data, output_path)
    # generate virtual datasets from slices
    if 'shape' in data.keys():
        layout = handle_shape_entries(data, file, path)
        grp.create_virtual_dataset(entry_name, layout)
    # multiple datasets to concatenate
    elif 'link' in data.keys() and isinstance(data['link'], list):
        total_length = 0
        sources = []
        for index, source_file in enumerate(file):
            vsource = h5py.VirtualSource(source_file,
                                         path[index],
                                         shape=h5py.File(source_file, 'r')[path[index]].shape)
            total_length += vsource.shape[0]
            sources.append(vsource)
        layout = h5py.VirtualLayout(shape=total_length, dtype=np.float64)
        offset = 0
        for vsource in sources:
            layout[offset:offset + vsource.shape[0]] = vsource
            offset += vsource.shape[0]
        grp.create_virtual_dataset(entry_name, layout, fillvalue=0)
    # internal and external links
    elif 'link' in data.keys():
        if ':/' not in data['link']:
            grp[entry_name] = h5py.SoftLink(path)  # internal link
        else:
            grp[entry_name] = h5py.ExternalLink(file, path)  # external link
    elif 'compress' in data.keys():
        if not (isinstance(data["compress"], str) or np.isscalar(data["compress"])):
            strength = 9  # strongest compression is space efficient but can take long
            accept = ("strength" in data.keys()) \
                and (isinstance(data['strength'], int)) \
                and (data["strength"] >= 0) \
                and (data["strength"] <= 9)
            if accept is True:
                strength = data["strength"]
            grp.create_dataset(entry_name,
                               data=data["compress"],
                               compression="gzip",
                               chunks=True,
                               compression_opts=strength)
        else:
            grp.create_dataset(entry_name, data=data["compress"])
    else:
        raise InvalidDictProvided("A dictionary was provided to the template but it didn't"
                                  " fall into any of the know cases of handling"
                                  " dictionaries. This occured for: " + entry_name)
    # Check whether link has been stabilished or not
    try:
        return grp[entry_name]
    except KeyError:
        logger.warning("No path '%s' available to be linked.", path)
        del grp[entry_name]
        return None


class Writer:
    """The writer class for writing a NeXus file in accordance with a given NXDL.

    Args:
        data (dict): Dictionary containing the data to convert.
        nxdl_path (str): Path to the nxdl file to use during conversion.
        output_path (str): Path to the output NeXus file.

    Attributes:
        data (dict): Dictionary containing the data to convert.
        nxdl_path (str): Path to the nxdl file to use during conversion.
        output_path (str): Path to the output NeXus file.
        output_nexus (h5py.File): The h5py file object to manipulate output file.
        nxdl_data (dict): Stores xml data from given nxdl file to use during conversion.
        nxs_namespace (str): The namespace used in the NXDL tags. Helps search for XML children.
    """

    def __init__(self, data: dict = None, nxdl_path: str = None, output_path: str = None):
        """Constructs the necessary objects required by the Writer class."""
        self.data = data
        self.nxdl_path = nxdl_path
        self.output_path = output_path
        self.output_nexus = h5py.File(self.output_path, "w")
        self.nxdl_data = ET.parse(self.nxdl_path).getroot()
        self.nxs_namespace = get_namespace(self.nxdl_data)

    def __nxdl_to_attrs(self, path: str = '/') -> dict:
        """
        Return a dictionary of all the attributes at the given path in the NXDL and
        the required attribute values that were requested in the NXDL from the data.

        If an NXDL attribute was not found, it returns None.
        """
        nxdl_path = helpers.convert_data_converter_dict_to_nxdl_path(path)

        try:
            elem = nexus.get_node_at_nxdl_path(
                nxdl_path, elem=copy.deepcopy(self.nxdl_data))
        except nexus.NxdlAttributeError:
            return None

        # Remove the name attribute as we only use it to name the HDF5 entry
        if "name" in elem.attrib.keys():
            del elem.attrib["name"]

        # Fetch values for required attributes requested by the NXDL
        for attr_name in elem.findall(f"{self.nxs_namespace}attribute"):
            key = f"{path}/@{attr_name.get('name')}"
            if key in self.data:
                elem.attrib[attr_name.get('name')] = self.data[key]

        return elem.attrib

    def ensure_and_get_parent_node(self, path: str, undocumented_paths) -> h5py.Group:
        """Returns the parent if it exists for a given path else creates the parent group."""
        parent_path = path[0:path.rindex('/')] or '/'
        parent_path_hdf5 = helpers.convert_data_dict_path_to_hdf5_path(parent_path)
        if not does_path_exist(parent_path, self.output_nexus):
            parent = self.ensure_and_get_parent_node(parent_path, undocumented_paths)
            grp = parent.create_group(parent_path_hdf5)

            attrs = self.__nxdl_to_attrs(parent_path)

            if attrs is not None:
                grp.attrs['NX_class'] = attrs["type"]
            return grp
        return self.output_nexus[parent_path_hdf5]

    def write(self):
        """Writes the NeXus file with previously validated data from the reader with NXDL attrs."""
        hdf5_links_for_later = []

        def add_units_key(dataset, path):
            units_key = f"{path}/@units"
            if units_key in self.data.keys() and self.data[units_key] is not None:
                dataset.attrs["units"] = self.data[units_key]

        for path, value in self.data.items():
            try:
                if path[path.rindex('/') + 1:] == '@units':
                    continue

                entry_name = helpers.get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
                if is_not_data_empty(value):
                    data = value
                else:
                    continue

                if entry_name[0] != "@":
                    grp = self.ensure_and_get_parent_node(path, self.data.undocumented.keys())
                    if isinstance(data, dict):
                        if "compress" in data.keys():
                            dataset = handle_dicts_entries(data, grp, entry_name,
                                                           self.output_path, path)
                        else:
                            hdf5_links_for_later.append([data, grp, entry_name,
                                                         self.output_path, path])
                    else:
                        dataset = grp.create_dataset(entry_name, data=data)
            except InvalidDictProvided as exc:
                print(str(exc))
            except Exception as exc:
                raise IOError(f"Unknown error occured writing the path: {path} "
                              f"with the following message: {str(exc)}") from exc
        for links in hdf5_links_for_later:
            dataset = handle_dicts_entries(*links)
            if dataset is None:
                # If target of a link is invalid to be linked
                del self.data[links[-1]]

        for path, value in self.data.items():
            try:
                if path[path.rindex('/') + 1:] == '@units':
                    continue

                entry_name = helpers.get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
                if is_not_data_empty(value):
                    data = value
                else:
                    continue

                if entry_name[0] != "@":
                    path_hdf5 = helpers.convert_data_dict_path_to_hdf5_path(path)

                    add_units_key(self.output_nexus[path_hdf5], path)
                else:
                    # consider changing the name here the lvalue can also be group!
                    dataset = self.ensure_and_get_parent_node(path, self.data.undocumented.keys())
                    dataset.attrs[entry_name[1:]] = data
            except Exception as exc:
                raise IOError(f"Unknown error occured writing the path: {path} "
                              f"with the following message: {str(exc)}") from exc

        self.output_nexus.close()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parser for adding default plots and path to them to template."""

# pylint: disable=no-member

import numpy as np

# ##MK::for EDS use the spectrum stack, the more generic, the more complex
# ##MK::the logic has to be to infer a default plottable
# when using hyperspy and EDS data, the path to the default plottable
# should point to an existent plot, in this example we use the
# NxSpectrumSetEmXray stack_data instance in the first event ...
# default plot selection preference in decreasing order of relevance
# ebsd > xray > eels > adf > bse, se
# spectrum_set_ebsd: ipf > overview > stack
# spectrum_set_xray: stack > summary > overview
# spectrum_set_eels: stack > summary > overview
# image_set_adf: overview
# image_set_se/bse: overview


def xray_plot_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance for Xray."""
    entry_name = f"entry{entry_id}"
    trg = f"/ENTRY[{entry_name}]/measurement/" \
          f"EVENT_DATA_EM[event_data_em1]/xray/"

    path = ""
    if f"{trg}stack/DATA[data_counts]" in template.keys():
        assert isinstance(
            template[f"{trg}stack/DATA[data_counts]"]["compress"], np.ndarray), \
            "EDS data stack not existent!"
        path = "stack"
    if f"{trg}summary/DATA[data_counts]" in template.keys():
        assert isinstance(
            template[f"{trg}summary/DATA[data_counts]"]["compress"], np.ndarray), \
            "EDS data summary not existent!"
        path = "summary"

    if path != "":
        print(f"Found xray default plot for {entry_name} at {path}")
        trg = "/"
        template[f"{trg}@default"] = entry_name
        trg = f"/ENTRY[{entry_name}]/"
        template[f"{trg}@default"] = "measurement"
        trg += "measurement/"
        template[f"{trg}@default"] = "event_data_em1"
        trg += "EVENT_DATA_EM[event_data_em1]/"
        template[f"{trg}@default"] = "xray"
        trg += "xray/"
        template[f"{trg}@default"] = path
        return True

    return False


def eels_plot_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance for EELS."""
    entry_name = f"entry{entry_id}"
    trg = f"/ENTRY[{entry_name}]/measurement/" \
          f"EVENT_DATA_EM[event_data_em1]/eels/"

    path = ""
    if f"{trg}stack/DATA[data_counts]" in template.keys():
        assert isinstance(
            template[f"{trg}stack/DATA[data_counts]"]["compress"], np.ndarray), \
            "EELS data stack not existent!"
        path = "stack"
    if f"{trg}summary/DATA[data_counts]" in template.keys():
        assert isinstance(
            template[f"{trg}summary/DATA[data_counts]"]["compress"], np.ndarray), \
            "EELS data summary not existent!"
        path = "summary"

    if path != "":
        print(f"Found eels default plot for {entry_name} at {path}")
        trg = "/"
        template[f"{trg}@default"] = entry_name
        trg = f"/ENTRY[{entry_name}]/"
        template[f"{trg}@default"] = "measurement"
        trg += "measurement/"
        template[f"{trg}@default"] = "event_data_em1"
        trg += "EVENT_DATA_EM[event_data_em1]/"
        template[f"{trg}@default"] = "eels"
        trg += "eels/"
        template[f"{trg}@default"] = path
        return True

    return False


def adf_plot_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance for ADF."""
    entry_name = f"entry{entry_id}"
    trg = f"/ENTRY[{entry_name}]/measurement/" \
          f"EVENT_DATA_EM[event_data_em1]/adf/"

    path = ""
    if f"{trg}stack/DATA[data_counts]" in template.keys():
        assert isinstance(
            template[f"{trg}stack/DATA[data_counts]"]["compress"], np.ndarray), \
            "ADF data stack not existent!"
        path = "stack"

    if path != "":
        print(f"Found adf default plot for entry {entry_name} at {path}")
        trg = "/"
        template[f"{trg}@default"] = entry_name
        trg = f"/ENTRY[{entry_name}]/"
        template[f"{trg}@default"] = "measurement"
        trg += "measurement/"
        template[f"{trg}@default"] = "event_data_em1"
        trg += "EVENT_DATA_EM[event_data_em1]/"
        template[f"{trg}@default"] = "adf"
        trg += "adf/"
        template[f"{trg}@default"] = path
        return True

    return False


def image_plot_available(template: dict, entry_id: int) -> bool:
    """Choose a preferred NXdata/data instance for generic image."""
    entry_name = f"entry{entry_id}"
    trg = f"/ENTRY[{entry_name}]/measurement/EVENT_DATA_EM[event_data_em1]/" \
          f"IMAGE_SET[image_set1]/"

    path = ""
    if f"{trg}DATA[stack]/data_counts" in template.keys():
        assert isinstance(
            template[f"{trg}DATA[stack]/data_counts"]["compress"], np.ndarray), \
            "Generic image data stack not existent!"
        path = "stack"

    if path != "":
        print(f"Found image default plot for {entry_name} at {path}")
        trg = "/"
        template[f"{trg}@default"] = entry_name
        trg = f"/ENTRY[{entry_name}]/"
        template[f"{trg}@default"] = "measurement"
        trg += "measurement/"
        template[f"{trg}@default"] = "event_data_em1"
        trg += "EVENT_DATA_EM[event_data_em1]/"
        template[f"{trg}@default"] = "image_set1"
        trg += "IMAGE_SET[image_set1]/"
        template[f"{trg}@default"] = path
        return True

    return False


def em_spctrscpy_default_plot_generator(template: dict, n_entries: int) -> dict:  # ignore:R0915
    """For a valid NXS file at least one default plot is required."""

    for entry_id in np.arange(1, n_entries + 1):
        # by convention showing always the first entry of an NXevent_data_em

        # add ebsd

        if xray_plot_available(template, entry_id) is True:
            continue

        if eels_plot_available(template, entry_id) is True:
            continue

        if adf_plot_available(template, entry_id) is True:
            continue

        # add bse, se

        if image_plot_available(template, entry_id) is True:
            continue

        print(f"WARNING: No path to a default plot found for entry{entry_id} !")

    return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""An example reader implementation for the DataConverter."""
from typing import Tuple, Any
import json
import pickle
import numpy as np
import xarray
from mergedeep import merge

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.template import Template
from pynxtools.dataconverter import hdfdict


def parse_slice(slice_string):
    """Converts slice strings to actual tuple sets of slices for index syntax"""
    slices = slice_string.split(",")
    for index, item in enumerate(slices):
        values = item.split(":")
        slices[index] = slice(*[None if x == '' else int(x) for x in values])
    return np.index_exp[tuple(slices)]


def get_val_nested_keystring_from_dict(keystring, data):
    """
    Fetches data from the actual data dict using path strings without a leading '/':
        'path/to/data/in/dict'
    """
    if isinstance(keystring, (list, dict)):
        return keystring

    current_key = keystring.split("/")[0]
    if isinstance(data[current_key], (dict, hdfdict.LazyHdfDict)):
        return get_val_nested_keystring_from_dict(keystring[keystring.find("/") + 1:],
                                                  data[current_key])
    if isinstance(data[current_key], xarray.DataArray):
        return data[current_key].values
    if isinstance(data[current_key], xarray.core.dataset.Dataset):
        raise NotImplementedError("Xarray datasets are not supported. "
                                  "You can only use xarray dataarrays.")

    return data[current_key]


def get_attrib_nested_keystring_from_dict(keystring, data):
    """
    Fetches all attributes from the data dict using path strings without a leading '/':
        'path/to/data/in/dict'
    """
    if isinstance(keystring, (list, dict)):
        return keystring

    key_splits = keystring.split("/")
    parents = key_splits[:-1]
    target = key_splits[-1]
    for key in parents:
        data = data[key]

    return data[target + "@"] if target + "@" in data.keys() else None


def is_path(keystring):
    """Checks whether a given value in the mapping is a mapping path or just data"""
    return isinstance(keystring, str) and len(keystring) > 0 and keystring[0] == "/"


def fill_undocumented(mapping, template, data):
    """Fill the extra paths provided in the map file that are not in the NXDL"""
    for path, value in mapping.items():
        if is_path(value):
            template["undocumented"][path] = get_val_nested_keystring_from_dict(value[1:],
                                                                                data)
            fill_attributes(path, value[1:], data, template)
        else:
            template["undocumented"][path] = value


def fill_documented(template, mapping, template_provided, data):
    """Fill the needed paths that are explicitly mentioned in the NXDL"""
    for req in ("required", "optional", "recommended"):
        for path in template_provided[req]:
            try:
                map_str = mapping[path]
                if is_path(map_str):
                    template[path] = get_val_nested_keystring_from_dict(map_str[1:],
                                                                        data)
                    fill_attributes(path, map_str[1:], data, template)
                else:
                    template[path] = map_str

                del mapping[path]
            except KeyError:
                pass


def fill_attributes(path, map_str, data, template):
    """Fills in the template all attributes found in the data object"""
    attribs = get_attrib_nested_keystring_from_dict(map_str, data)
    if attribs:
        for key, value in attribs.items():
            template[path + "/@" + key] = value


def convert_shapes_to_slice_objects(mapping):
    """Converts shape slice strings to slice objects for indexing"""
    for key in mapping:
        if isinstance(mapping[key], dict):
            if "shape" in mapping[key]:
                mapping[key]["shape"] = parse_slice(mapping[key]["shape"])


def get_map_from_partials(partials, template, data):
    """Takes a list of partials and returns a mapping dictionary to fill partials in our template"""
    mapping: dict = {}
    for partial in partials:
        path = ""
        template_path = ""
        for part in partial.split("/")[1:]:
            path = path + "/" + part
            attribs = get_attrib_nested_keystring_from_dict(path[1:], data)
            if template_path + "/" + part in template.keys():
                template_path = template_path + "/" + part
            else:
                nx_name = f"{attribs['NX_class'][2:].upper()}[{part}]" if attribs and "NX_class" in attribs else part  # pylint: disable=line-too-long
                template_path = template_path + "/" + nx_name
        mapping[template_path] = path

    return mapping


class JsonMapReader(BaseReader):
    """A reader that takes a mapping json file and a data file/object to return a template."""

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXtest", "*"]

    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """
        Reads data from given file and returns a filled template dictionary.

        Only the data object is expected to be passed as the first object.
        Alteratively, a data object represented with a file.json or file.xarray.pickle
        can also be used.
        The mapping is only accepted as file.mapping.json to the inputs.
        """
        data: dict = {}
        mapping: dict = None
        partials: list = []

        data = objects[0] if objects else data

        for file_path in file_paths:
            file_extension = file_path[file_path.rindex("."):]
            if file_extension == ".json":
                with open(file_path, "r", encoding='utf-8') as input_file:
                    if ".mapping" in file_path:
                        mapping = json.loads(input_file.read())
                    else:
                        data = json.loads(input_file.read())
            elif file_extension == ".pickle":
                with open(file_path, "rb") as input_file:  # type: ignore[assignment]
                    data = pickle.load(input_file)  # type: ignore[arg-type]
            else:
                is_hdf5 = False
                with open(file_path, "rb") as input_file:
                    if input_file.read(8) == b'\x89HDF\r\n\x1a\n':
                        is_hdf5 = True
                if is_hdf5:
                    hdf = hdfdict.load(file_path)
                    hdf.unlazy()
                    merge(data, dict(hdf))
                    if "entry@" in data and "partial" in data["entry@"]:
                        partials.extend(data["entry@"]["partial"])

        if mapping is None:
            if len(partials) > 0:
                mapping = get_map_from_partials(partials, template, data)
            else:
                template = Template({x: "/hierarchical/path/in/your/datafile" for x in template})
                raise IOError("Please supply a JSON mapping file: --input-file"
                              " my_nxdl_map.mapping.json\n\n You can use this "
                              "template for the required fields: \n" + str(template))

        new_template = Template()
        convert_shapes_to_slice_objects(mapping)

        fill_documented(new_template, mapping, template, data)

        fill_undocumented(mapping, new_template, data)

        return new_template


# This has to be set to allow the convert script to use this reader. Set it to "MyDataReader".
READER = JsonMapReader



"""init file

"""
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re
from datetime import datetime
from glob import glob
from typing import Union

from pynxtools._build_wrapper import get_vcs_version
from pynxtools.definitions.dev_tools.globals.nxdl import get_nxdl_version

MAIN_BRANCH_NAME = "fairmat"


def _build_version(tag: str, distance: int, node: str, dirty: bool) -> str:
    """
    Builds the version string for a given set of git states.
    This resembles `no-guess-dev` + `node-and-date` behavior from setuptools_scm.
    """
    if distance == 0 and not dirty:
        return f"{tag}"

    dirty_appendix = datetime.now().strftime(".d%Y%m%d") if dirty else ""
    return f"{tag}.post1.dev{distance}+{node}{dirty_appendix}"


def format_version(version: str) -> str:
    """
    Formats the git describe version string into the local format.
    """
    version_parts = version.split("-")

    return _build_version(
        version_parts[0],
        int(version_parts[1]),
        version_parts[2],
        len(version_parts) == 4 and version_parts[3] == "dirty",
    )


def get_nexus_version() -> str:
    """
    The version of the Nexus standard and the NeXus Definition language
    based on git tags and commits
    """
    version = get_vcs_version()

    if version is not None:
        return format_version(version)

    version_file = os.path.join(os.path.dirname(__file__), "nexus-version.txt")

    if not os.path.exists(version_file):
        # We are in the limbo, just get the nxdl version from nexus definitions
        return format_version(get_nxdl_version())

    with open(version_file, encoding="utf-8") as vfile:
        return format_version(vfile.read().strip())


def get_nexus_version_hash() -> str:
    """
    Gets the git hash from the nexus version string
    """
    version = re.search(r"g([a-z0-9]+)", get_nexus_version())

    if version is None:
        return MAIN_BRANCH_NAME

    return version.group(1)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""MPES reader implementation for the DataConverter."""
import errno
import json
import os
from functools import reduce
from typing import Any
from typing import Tuple

import h5py
import xarray as xr
import yaml

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.readers.utils import flatten_and_replace, FlattenSettings

DEFAULT_UNITS = {
    "X": "step",
    "Y": "step",
    "t": "step",
    "tofVoltage": "V",
    "extractorVoltage": "V",
    "extractorCurrent": "A",
    "cryoTemperature": "K",
    "sampleTemperature": "K",
    "dldTimeBinSize": "ns",
    "delay": "ps",
    "timeStamp": "s",
    "energy": "eV",
    "kx": "1/A",
    "ky": "1/A",
}


def res_to_xarray(res, bin_names, bin_axes, metadata=None):
    """creates a BinnedArray (xarray subclass) out of the given np.array
    Parameters:
        res: np.array
            nd array of binned data
        bin_names (list): list of names of the binned axes
        bin_axes (list): list of np.arrays with the values of the axes
    Returns:
        ba: BinnedArray (xarray)
            an xarray-like container with binned data, axis, and all
            available metadata
    """
    dims = bin_names
    coords = {}
    for name, vals in zip(bin_names, bin_axes):
        coords[name] = vals

    xres = xr.DataArray(res, dims=dims, coords=coords)

    for name in bin_names:
        try:
            xres[name].attrs["unit"] = DEFAULT_UNITS[name]
        except KeyError:
            pass

    xres.attrs["units"] = "counts"
    xres.attrs["long_name"] = "photoelectron counts"

    if metadata is not None:
        xres.attrs["metadata"] = metadata

    return xres


def h5_to_xarray(faddr, mode="r"):
    """Rear xarray data from formatted hdf5 file
    Args:
        faddr (str): complete file name (including path)
        mode (str): hdf5 read/write mode
    Returns:
        xarray (xarray.DataArray): output xarra data
    """
    with h5py.File(faddr, mode) as h5_file:
        # Reading data array
        try:
            data = h5_file["binned"]["BinnedData"]
        except KeyError:
            print("Wrong Data Format, data not found")
            raise

        # Reading the axes
        axes = []
        bin_names = []

        try:
            for axis in h5_file["axes"]:
                axes.append(h5_file["axes"][axis])
                bin_names.append(h5_file["axes"][axis].attrs["name"])
        except KeyError:
            print("Wrong Data Format, axes not found")
            raise

        # load metadata
        if "metadata" in h5_file:

            def recursive_parse_metadata(node):
                if isinstance(node, h5py.Group):
                    dictionary = {}
                    for key, value in node.items():
                        dictionary[key] = recursive_parse_metadata(value)

                else:
                    dictionary = node[...]
                    try:
                        dictionary = dictionary.item()
                        if isinstance(dictionary, (bytes, bytearray)):
                            dictionary = dictionary.decode()
                    except ValueError:
                        pass

                return dictionary

            metadata = recursive_parse_metadata(h5_file["metadata"])
        # Segment to change Vset to V in lens voltages
        if "file" in metadata.keys():
            for k in list(metadata['file']):
                if "VSet" in k:
                    key = k[:-3]
                    metadata['file'][key] = metadata['file'][k]
                    del metadata['file'][k]

        xarray = res_to_xarray(data, bin_names, axes, metadata)
        return xarray


def iterate_dictionary(dic, key_string):
    """Recursively iterate in dictionary and give back its values"""
    keys = key_string.split("/", 1)
    if keys[0] in dic:
        if len(keys) == 1:
            return dic[keys[0]]
        if not len(keys) == 1:
            return iterate_dictionary(dic[keys[0]], keys[1])
    else:
        raise KeyError
    return None


CONVERT_DICT = {
    'Instrument': 'INSTRUMENT[instrument]',
    'Analyzer': 'ELECTRONANALYSER[electronanalyser]',
    'Manipulator': 'MANIPULATOR[manipulator]',
    'Beam': 'BEAM[beam]',
    'unit': '@units',
    'Sample': 'SAMPLE[sample]',
    'Source': 'SOURCE[source]',
    'User': 'USER[user]'
}

REPLACE_NESTED = {
    'SOURCE[source]/Probe': 'SOURCE[source]',
    'SOURCE[source]/Pump': 'SOURCE[source_pump]',
    'BEAM[beam]/Probe': 'BEAM[beam]',
    'BEAM[beam]/Pump': 'BEAM[beam_pump]',
    'sample_history': 'sample_history/description'
}


def handle_h5_and_json_file(file_paths, objects):
    """Handle h5 or json input files."""
    x_array_loaded = xr.DataArray()
    config_file_dict = {}
    eln_data_dict = {}

    for file_path in file_paths:
        try:
            file_extension = file_path[file_path.rindex("."):]
        except ValueError as exc:
            raise ValueError(
                f"The file path {file_path} must have an extension.",
            ) from exc

        extentions = [".h5", ".json", ".yaml", ".yml"]
        if file_extension not in extentions:
            print(
                f"WARNING \n"
                f"The reader only supports files of type {extentions}, "
                f"but {file_path} does not match.",
            )

        if not os.path.exists(file_path):
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                file_path,
            )

        if file_extension == ".h5":
            x_array_loaded = h5_to_xarray(file_path)
        elif file_extension == ".json":
            with open(file_path, encoding="utf-8") as file:
                config_file_dict = json.load(file)
        elif file_extension in [".yaml", ".yml"]:
            with open(file_path, encoding="utf-8") as feln:
                eln_data_dict = flatten_and_replace(
                    FlattenSettings(
                        dic=yaml.safe_load(feln),
                        convert_dict=CONVERT_DICT,
                        replace_nested=REPLACE_NESTED
                    )
                )

    if objects is not None:
        # For the case of a single object
        assert isinstance(
            objects,
            xr.core.dataarray.DataArray,
        ), "The given object must be an xarray"
        x_array_loaded = objects

    return x_array_loaded, config_file_dict, eln_data_dict


def rgetattr(obj, attr):
    """Get attributes recursively"""
    def _getattr(obj, attr):
        return getattr(obj, attr)

    if "index" in attr:
        axis = attr.split(".")[0]
        return obj.dims.index(f"{axis}")

    return reduce(_getattr, [obj] + attr.split("."))


def fill_data_indices_in_config(config_file_dict, x_array_loaded):
    """Add data indices key value pairs to the config_file
    dictionary from the xarray dimensions if not already
    present.
    """
    for key in list(config_file_dict):
        if "*" in key:
            value = config_file_dict[key]
            for dim in x_array_loaded.dims:
                new_key = key.replace("*", dim)
                new_value = value.replace("*", dim)

                if new_key not in config_file_dict.keys() \
                        and new_value not in config_file_dict.values():
                    config_file_dict[new_key] = new_value

            config_file_dict.pop(key)


class MPESReader(BaseReader):
    """MPES-specific reader class"""

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXmpes"]

    def read(  # pylint: disable=too-many-branches
            self,
            template: dict = None,
            file_paths: Tuple[str] = None,
            objects: Tuple[Any] = None,
    ) -> dict:
        """Reads data from given file or alternatively an xarray object
        and returns a filled template dictionary"""

        if not file_paths:
            raise IOError("No input files were given to MPES Reader.")

        (
            x_array_loaded,
            config_file_dict,
            eln_data_dict,
        ) = handle_h5_and_json_file(file_paths, objects)

        fill_data_indices_in_config(config_file_dict, x_array_loaded)

        for key, value in config_file_dict.items():

            if isinstance(value, str) and ":" in value:
                precursor = value.split(":")[0]
                value = value[value.index(":") + 1:]

                # Filling in the data and axes along with units from xarray
                if precursor == "@data":
                    try:
                        template[key] = rgetattr(
                            obj=x_array_loaded,
                            attr=value,
                        )
                        if key.split("/")[-1] == "@axes":
                            template[key] = list(template[key])

                    except ValueError:
                        print(
                            f"Incorrect axis name corresponding to "
                            f"the path {key}",
                        )

                    except AttributeError:
                        print(
                            f"Incorrect naming syntax or the xarray doesn't "
                            f"contain entry corresponding to the path {key}",
                        )

                # Filling in the metadata from xarray
                elif precursor == "@attrs":
                    if key not in eln_data_dict:
                        try:  # Tries to fill the metadata
                            template[key] = iterate_dictionary(
                                x_array_loaded.attrs,
                                value,
                            )

                        except KeyError:
                            print(
                                f"[info]: Path {key} not found. "
                                f"Skipping the entry.",
                            )

            else:
                # Fills in the fixed metadata
                template[key] = value

        # Filling in ELN metadata and overwriting the common paths by
        # giving preference to the ELN metadata
        for key, value in eln_data_dict.items():
            template[key] = value

        return template


READER = MPESReader



# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=too-many-lines

"""
Class for reading XPS files from TXT export of Scienta.
"""

from datetime import datetime
import pytz
import numpy as np


class ScientaTxtParser():
    """ Parser for Scienta TXT exports."""

    # pylint: disable=too-few-public-methods

    def __init__(self):
        self.lines: list = []
        self.spectra: list = []
        self.no_of_regions = 0

        keys_map = {
            'Number of Regions': 'no_of_regions',
            'Version': 'software_version',
            'Lens Mode': 'lens_mode',
            'Pass Energy': 'pass_energy',
            'Number of Sweeps': 'no_of_scans',
            'Excitation Energy': 'excitation_energy',
            'Energy Scale': 'energy_scale',
            'Acquisition Mode': 'acquisition_mode',
            'Energy Unit': 'x_units',
            'Center Energy': 'center_energy',
            'Low Energy': 'start_energy',
            'High Energy': 'stop_energy',
            'Energy Step': 'step_size',
            'Step Time': 'dwell_time',
            'Number of Slices': 'number_of_slices',
            'File': 'data_file',
            'Sequence': 'sequence_file',
            'Spectrum Name': 'spectrum_type',
            'Instrument': 'instrument_name',
            'Location': 'vendor',
            'User': 'user_initials',
            'Sample': 'sample_name',
            'Comments': 'spectrum_comment',
            'Date': 'start_date',
            'Time': 'start_time',
            'Time per Spectrum Channel': 'time_per_spectrum_channel',
            'DetectorMode': 'detector_mode',
        }

        self.region_keys_map = {
            'Region Name': 'region_name',
            'name': 'energy_type',
            'size': 'energy_size',
            'scale': 'energy_axis',
        }

        detector_map = {
            'Detector First X-Channel': 'detector_first_x_channel',
            'Detector Last X-Channel': 'detector_last_x_channel',
            'Detector First Y-Channel': 'detector_first_y_channel',
            'Detector Last Y-Channel': 'detector_last_y_channel',
        }

        lens_mode_map = {
            'Transmission': 'fixed analyzer transmission'
        }

        self.key_maps = [
            keys_map,
            self.region_keys_map,
            detector_map,
            lens_mode_map,
        ]

        self.value_map = {
            'no_of_regions': self._change_no_of_regions,
            'x_units': self._change_energy_type,
            'energy_axis': [self._separate_dimension_scale, np.array],
            'energy_scale': self._change_energy_type,
            'acquisition_mode': self._change_scan_mode,
        }

    def parse_file(self, file):
        """
        Parse the file's data and metadata into a flat
        list of dictionaries.


        Parameters
        ----------
        file : str
            Filepath of the TXT file to be read.

        Returns
        -------
        self.spectra
            Flat list of dictionaries containing one spectrum each.

        """
        self._read_lines(file)
        self._parse_header()

        for region_id in range(1, self.no_of_regions + 1):
            self._parse_region(region_id)

        return self.spectra

    def _read_lines(self, filepath):
        """
        Read all lines from the input txt files.


        Parameters
        ----------
        filepath : str
            Filepath of the TXT file to be read.

        Returns
        -------
        None.

        """
        with open(filepath, encoding="utf-8") as file:
            for line in file:
                self.lines += [line]

    def _parse_header(self):
        """
        Parse header with information about the software version
        and the number of spectra in the file.

        Returns
        -------
        None.

        """
        n_headerlines = 4
        header = self.lines[:n_headerlines]
        self.lines = self.lines[n_headerlines:]

        for line in header:
            key, value = self._get_key_value_pair(line)
            if key:
                value = self._re_map_values(key, value)
                setattr(self, key, value)

    def _parse_region(self, region_id):
        """
        Parse data from one region (i.e., one measured spectrum)
        into a dictioanry and append to all spectra.

        Parameters
        ----------
        region_id : int
            Number of the region in the file.

        Returns
        -------
        None.

        """

        region_data = {
            'region_id': region_id,
            'data': {
                'x': [],
                'y': []
            },
        }

        begin_region = False
        begin_info = False
        begin_data = False

        for line in self.lines:
            if line.startswith(f'[Region {region_id}'):
                begin_region = True
            if line.startswith(f'[Info {region_id}'):
                begin_info = True
            if line.startswith(f'[Data {region_id}'):
                begin_data = True
                continue

            if line.startswith("\n"):
                begin_region = False
                begin_info = False
                begin_data = False

            if begin_region:
                # Read region meta data.
                key, value = self._get_key_value_pair(line)
                if 'Dimension' in key:
                    key = self._re_map_keys(key)
                    key = self.region_keys_map[key.rsplit(' ')[-1]]
                    key = key.rsplit(' ')[-1]
                    value = self._re_map_values(key, value)
                    if self._check_valid_value(value):
                        region_data[key] = value

            if begin_info:
                # Read instrument meta data for this region.
                key, value = self._get_key_value_pair(line)
                if self._check_valid_value(value):
                    region_data[key] = value

            if begin_data:
                # Read XY data for this region.
                [energy, intensity] = \
                    [float(s) for s in line.split(' ') if s != '']

                region_data['data']['x'].append(energy)
                region_data['data']['y'].append(intensity)

        region_data['data']['x'] = np.array(region_data['data']['x'])
        region_data['data']['y'] = np.array(region_data['data']['y'])

        # Convert date and time to ISO8601 date time.
        start_date = region_data['start_date']
        start_time = region_data['start_time']
        region_data['time_stamp'] = self._construct_date_time(
            start_date,
            start_time)

        self.spectra.append(region_data)

    def _check_valid_value(self, value):
        """
        Check if a string or an array is empty.

        Parameters
        ----------
        value : obj
            For testing, this can be a str or a np.ndarray.

        Returns
        -------
        bool
            True if the string or np.ndarray is not empty.

        """
        if isinstance(value, str) and value:
            return True
        if isinstance(value, np.ndarray) and value.size != 0:
            return True
        return False

    def _get_key_value_pair(self, line):
        """
        Split the line at the '=' sign and return a
        key-value pair. The values are mapped according
        to the desired format.

        Parameters
        ----------
        line : str
            One line from the input file.

        Returns
        -------
        key : str
            Anything before the '=' sign, mapped to the desired
            key format.
        value : obj
            Anything after the '=' sign, mapped to the desired
            value format and type.

        """
        try:
            [key, value] = line.split('=')
            key = self._re_map_keys(key)
            value = self._re_map_values(key, value)
            return key, value

        except ValueError:
            key, value = '', ''
            return key, value

    def _re_map_keys(self, key):
        """
        Map the keys returned from the file to the preferred keys for
        the parser output.

        """
        maps = {}
        for key_map in self.key_maps:
            maps.update(key_map)

        keys = list(maps.keys())

        if key in keys:
            key = maps[key]

        return key

    def _change_no_of_regions(self, no_of_regions):
        """
        Make no_of_regions an integer.
        Maps e.g. from '0001' string to 1.

        Parameters
        ----------
        no_of_regions : str
            No. of regions in the data file,
            read from the first line.

        Returns
        -------
        int
            Ineger value of the number of regions.

        """
        return int(no_of_regions)

    def _change_energy_type(self, energy):
        """
        Change the strings for energy type to the preferred format.

        """
        if 'Binding' in energy:
            return 'binding energy'
        if 'Kinetic' in energy:
            return 'kinetic energy'
        return None

    def _separate_dimension_scale(self, scale):
        return [float(s) for s in scale.split(' ')]

    def _change_scan_mode(self, acquisition_mode):
        """
        Change the strings for acquisition mode type to
        the allowed values in NXmpes.

        """
        if acquisition_mode == 'Fixed':
            return 'fixed'
        if acquisition_mode == 'Swept':
            return 'sweep'
        return None

    def _construct_date_time(self, date, time):
        """
        Convert the native time format to the datetime string
        in the ISO 8601 format: '%Y-%b-%dT%H:%M:%S.%fZ'.

        """
        date_time = datetime.combine(
            datetime.strptime(date, "%Y-%m-%d"),
            datetime.strptime(time, '%H:%M:%S').time())

        localtz = pytz.timezone('Europe/Berlin')
        date_time = localtz.localize(date_time)

        return date_time.isoformat()

    def _re_map_values(self, input_key, value):
        """
        Map the values returned from the file to the preferred format for
        the parser output.

        """
        try:
            value = value.rstrip('\n')
        except AttributeError:
            pass

        keys = list(self.value_map.keys())

        for k in keys:
            if k in input_key:
                if not isinstance(self.value_map[k], list):
                    map_methods = [self.value_map[k]]
                else:
                    map_methods = self.value_map[k]
                for method in map_methods:
                    value = method(value)
        return value
